from os import path
import cStringIO
import string

from pysces import model, PyscesModel

from ._paths import get_model_name


__all__ = ['psc_to_str',
           'mod_to_str',
           'strip_fixed',
           'augment_fix_sting',
           'fix_metabolite',
           'fix_metabolite_ss']


def psc_to_str(name):
    """
    Takes a filename and returns a path of where this file should be found.

    Parameters
    ----------
    name : str
        A string containing a filename.

    Returns
    -------
    str
        A string indicating the path to a psc file.
    """
    if name[-4:] != '.psc':
        name += '.psc'
    F = file(path.join(PyscesModel.MODEL_DIR, name), 'r')
    fstr = F.read()
    F.close()
    return fstr


def mod_to_str(mod):
    """
    Converts an instantiated PySCeS model to a string.

    Parameters
    ----------
    mod : PysMod
        A Pysces model.

    Returns
    -------
    str
        A string representation of the contents of a PySCeS model file.

    """
    F = cStringIO.StringIO()
    mod.showModel(filename=F)
    fstr = F.getvalue()
    F.close()
    return fstr


def strip_fixed(fstr):
    """
    Take a psc file string and return two strings: (1) The file header
    containing the "FIX: " line and (2) the remainder of file.

    Parameters
    ----------
    fstr : str
        String representation of psc file.

    Returns
    -------
    tuple of str
        1st element contains file header, second the remainder of the file.

    See also
    --------
    psc_to_str
    mod_to_str
    """
    Fi = cStringIO.StringIO()
    Fi.write(fstr)
    Fi.seek(0)
    Fo = cStringIO.StringIO()
    Fhead = None
    for line in Fi:
        if line[:4] == "FIX:":
            Fhead = string.strip(line)
            Fo.write('\n')
        else:
            Fo.write(line)
    Fo.seek(0)
    return Fhead, Fo.read()


def augment_fix_sting(OrigFix, fix):
    """
    Adds a species to a psc file header.

    Parameters
    ----------
    OrigFix : str
        A psc file header
    fix : str
        Additional species to add to psc file header.

    Returns
    -------
    str
        A new psc file header that contains the contents of the original
        together with the new fixed species.
    """
    return OrigFix + ' %s' % fix


def fix_metabolite(mod, fix, model_name=None):
    """
    Fix a metabolite in a model and return a new model with the fixed
    metabolite.

    Parameters
    ----------
    mod : PysMod
        The original model.
    fix : str
        The metabolite to fix.
    model_name : str, optional (Default : none)
        The file name to use when saving the model (in psc/orca).
        If None it defaults to original_model_name_fix.

    Returns
    -------
    PysMod
        A new model instance with an additional fixed species.
    """
    assert fix in mod.species, "\nInvalid fixed species."

    if model_name is None:
        model_name = get_model_name(mod) + '_' + fix

    mod_str = mod_to_str(mod)
    fix_head, mod_str_sans_fix = strip_fixed(mod_str)
    new_fix_head = augment_fix_sting(fix_head, fix)
    new_mod = model(model_name, loader="string", fString=new_fix_head
                    + '\n' + mod_str_sans_fix)
    return new_mod


def fix_metabolite_ss(mod, fix, model_name=None):
    """
    Fix a metabolite at its steady state in a model and return a new
    model with the fixed metabolite.

    Parameters
    ----------
    mod : PysMod
        The original model.
    fix : str
        The metabolite to fix.
    model_name : str, optional (Default : none)
        The file name to use when saving the model (in psc/orca).
        If None it defaults to original_model_name_fix.

    Returns
    -------
    PysMod
        A new model instance with an additional fixed species.

    See Also
    --------
    fix_metabolite
    """
    mod.doState()
    fixed_ss = getattr(mod,fix)
    fix_mod = fix_metabolite(mod,fix,model_name)
    setattr(fix_mod,fix,fixed_ss)
    return fix_mod


import cPickle as pickle
from sympy.matrices import Matrix
from sympy import sympify
import sys

from ...utils.misc import extract_model
from ...utils.misc import get_filename_from_caller
from ...modeltools import make_path, get_file_path
from ...latextools import LatexExpr
from .symca_toolbox import SymcaToolBox as SMCAtools
from numpy import savetxt, array
from ...utils import ConfigReader
import warnings


all = ['Symca']

class Symca(object):
    """
    A class that performs Symbolic Metabolic Control Analysis.

    This class takes pysces model as an input and performs symbolic inversion
    of the ``E matrix`` using ``Sympy`` by calculating the determinant and
    adjoint matrices of this ``E matrix``.

    Parameters
    ----------
    mod : PysMod
        The pysces model on which to perform symbolic control analysis.
    auto_load : boolean
        If true

    Returns
    ------
    """

    def __init__(self, mod, auto_load=False, internal_fixed=False, ignore_steady_state=False, keep_zero_elasticities=True):
        super(Symca, self).__init__()
        ConfigReader.get_config()

        self._ignore_steady_state = ignore_steady_state
        self._keep_zero_ecs = keep_zero_elasticities
        self.mod, obj_type = extract_model(mod)
        if not self._ignore_steady_state:
            self.mod.doMca()
        else:
            warnings.warn_explicit("\nIgnoring steady-state solution: Steady-state variables set to 1. Note that parameter scan functionality is unavailable.",
                                   Warning,
                                   filename=get_filename_from_caller(),
                                   lineno=36)
            SMCAtools.populate_with_fake_elasticities(mod)
            SMCAtools.populate_with_fake_fluxes(mod)
            SMCAtools.populate_with_fake_ss_concentrations(mod)

        self._analysis_method = 'symca'
        self._internal_filename = 'object_data'
        self._working_dir = make_path(self.mod, self._analysis_method)
        self._ltxe = LatexExpr(self.mod)

        self.cc_results = None

        self._nmatrix = None
        self._species = None
        self._num_ind_species = None
        self._species_independent = None
        self._species_dependent = None
        self._fluxes = None
        self._num_ind_fluxes = None
        self._fluxes_independent = None
        self._fluxes_dependent = None
        self._kmatrix = None
        self._lmatrix = None
        self._subs_fluxes = None
        self._scaled_k = None
        self._scaled_l = None
        self._scaled_k0 = None
        self._scaled_l0 = None
        self._es_matrix = None
        self._esL = None
        self._ematrix = None

        self.internal_fixed = internal_fixed
        if obj_type == 'RateCharData':
            self.internal_fixed = True
        if auto_load:
            try:
                self.load_session()
            except:
                print 'Nothing to load_session: Run `do_symca` first'

    @property
    def nmatrix(self):
        if not self._nmatrix:
            self._nmatrix = SMCAtools.get_nmatrix(self.mod)

        return self._nmatrix

    @property
    def num_ind_species(self):
        if not self._num_ind_species:
            self._num_ind_species = SMCAtools.get_num_ind_species(self.mod)

        return self._num_ind_species

    @property
    def species(self):
        if not self._species:
            self._species = SMCAtools.get_species_vector(self.mod)

        return self._species

    @property
    def species_independent(self):
        if not self._species_independent:
            self._species_independent = Matrix(
                self.species[:self.num_ind_species]
            )

        return self._species_independent

    @property
    def species_dependent(self):
        if not self._species_dependent:
            self._species_dependent = Matrix(
                self.species[self.num_ind_species:]
            )

        return self._species_dependent

    @property
    def num_ind_fluxes(self):
        if not self._num_ind_fluxes:
            self._num_ind_fluxes = SMCAtools.get_num_ind_fluxes(self.mod)

        return self._num_ind_fluxes

    @property
    def fluxes(self):
        if not self._fluxes:
            self._fluxes = SMCAtools.get_fluxes_vector(self.mod)

        return self._fluxes

    @property
    def fluxes_independent(self):
        if not self._fluxes_independent:
            self._fluxes_independent = Matrix(
                self.fluxes[:self.num_ind_fluxes]
            )

        return self._fluxes_independent

    @property
    def fluxes_dependent(self):
        if not self._fluxes_dependent:
            self._fluxes_dependent = Matrix(
                self.fluxes[self.num_ind_fluxes:]
            )

        return self._fluxes_dependent

    @property
    def kmatrix(self):
        if not self._kmatrix:
            self._kmatrix = Matrix(self.mod.kmatrix)

        return self._kmatrix

    @property
    def lmatrix(self):
        if not self._lmatrix:
            self._lmatrix = Matrix(self.mod.lmatrix)

        return self._lmatrix

    @property
    def subs_fluxes(self):
        if not self._subs_fluxes:
            self._subs_fluxes = SMCAtools.substitute_fluxes(
                self.fluxes,
                self.kmatrix
            )

        return self._subs_fluxes

    @property
    def scaled_l(self):
        if not self._scaled_l:
            self._scaled_l = SMCAtools.scale_matrix(
                self.species,
                self.lmatrix,
                self.species_independent
            )

        return self._scaled_l

    @property
    def scaled_k(self):
        if not self._scaled_k:
            self._scaled_k = SMCAtools.scale_matrix(
                self.subs_fluxes,
                self.kmatrix,
                self.fluxes_independent
            )
        return self._scaled_k

    @property
    def scaled_l0(self):
        if not self._scaled_l0:
            self._scaled_l0 = self.scaled_l[self.num_ind_species:, :]

        return self._scaled_l0

    @property
    def scaled_k0(self):
        if not self._scaled_k0:
            self._scaled_k0 = self.scaled_k[self.num_ind_fluxes:, :]

        return self._scaled_k0

    @property
    def es_matrix(self):
        if not self._es_matrix:
            if self._ignore_steady_state or self._keep_zero_ecs:
                es_method = SMCAtools.get_es_matrix_no_mca
            else:
                es_method = SMCAtools.get_es_matrix
            self._es_matrix = es_method(
                self.mod,
                self.nmatrix,
                self.fluxes,
                self.species
            )

        return self._es_matrix

    @property
    def esL(self):
        if not self._esL:
            self._esL = self.es_matrix * self.scaled_l

        return self._esL

    @property
    def ematrix(self):
        if not self._ematrix:
            self._ematrix = SMCAtools.simplify_matrix(
                self.scaled_k.row_join(
                    -self.esL
                )
            )

        return self._ematrix

    def path_to(self, path):
        full_path = make_path(self.mod, self._analysis_method, [path])
        return full_path

    def save_session(self, file_name=None):
        file_name = get_file_path(working_dir=self._working_dir,
                                  internal_filename=self._internal_filename,
                                  fmt='pickle',
                                  file_name=file_name,
                                  write_suffix=False)

        assert self.cc_results, 'Nothing to save_session, run ``do_symca`` method first'
        main_cc_dict = SMCAtools.make_inner_dict(self.cc_results, 'cc_results')
        counter = 0
        while True:
            cc_container_name = 'cc_results_{0}'.format(counter)
            try:
                cc_container = getattr(self, cc_container_name)
                main_cc_dict.update(
                    SMCAtools.make_inner_dict(cc_container, cc_container_name))
                counter += 1
            except:
                break

        to_save = main_cc_dict
        with open(file_name, 'w') as f:
            pickle.dump(to_save, f)

    def load_session(self, file_name=None):
        file_name = get_file_path(working_dir=self._working_dir,
                                  internal_filename=self._internal_filename,
                                  fmt='pickle',
                                  file_name=file_name,
                                  write_suffix=False)

        with open(file_name, 'r') as f:
            main_cc_dict = pickle.load(f)

        cc_containers = {}
        for key, value in main_cc_dict.iteritems():
            common_denom_exp = value.pop('common_denominator')
            cc_container = SMCAtools.spawn_cc_objects(self.mod,
                                                      value.keys(),
                                                      [exp for exp in
                                                       value.values()],
                                                      common_denom_exp,
                                                      self._ltxe)
            cc_containers[key] = SMCAtools.make_CC_dot_dict(cc_container)
        for key, value in cc_containers.iteritems():
            setattr(self, key, value)

    def save_results(self, file_name=None, separator=',',fmt='%.9f'):
        file_name = get_file_path(working_dir=self._working_dir,
                                  internal_filename='cc_summary',
                                  fmt='csv',
                                  file_name=file_name, )

        rows = []
        cc_counter = 0
        cc_dicts = [self.cc_results]
        max_len = 0

        while True:
            try:
                next_dict = getattr(self, 'cc_results_%s' % cc_counter)
                cc_dicts.append(next_dict)
                cc_counter += 1
            except:
                break

        sep = ('######################', 0, '', '')
        cc_counter = -1
        for cc_dict in cc_dicts:
            result_name = '# results from cc_results'
            if cc_counter >= 0:
                result_name += '_%s' % cc_counter
            head = (result_name, 0, '', '')
            rows.append(head)
            for cc_name in sorted(cc_dict.keys()):
                cc_obj = cc_dict[cc_name]

                row_1 = (cc_obj.name,
                         cc_obj.value,
                         cc_obj.latex_name,
                         cc_obj.latex_expression)

                expr_len = len(cc_obj.latex_expression)
                if expr_len > max_len:
                    max_len = expr_len
                rows.append(row_1)
                if not cc_obj.name == 'common_denominator':
                    for cp in cc_obj.control_patterns.itervalues():
                        cols = (cp.name,
                                cp.value,
                                cp.latex_name,
                                cp.latex_expression)
                        rows.append(cols)
            rows.append(sep)
            cc_counter += 1

        str_fmt = 'S%s' % max_len
        head = ['name', 'value', 'latex_name', 'latex_expression']
        X = array(rows,
                  dtype=[(head[0], str_fmt),
                         (head[1], 'float'),
                         (head[2], str_fmt),
                         (head[3], str_fmt)])

        try:
            savetxt(fname=file_name,
                    X=X,
                    header=separator.join(head),
                    delimiter=separator,
                    fmt=['%s', fmt, '%s', '%s'],)

        except IOError as e:
            print e.strerror

    def do_symca(self, internal_fixed=None, auto_save_load=False):
        if internal_fixed is None:
            internal_fixed = self.internal_fixed

        def do_symca_internals(self):
            CC_i_num, common_denom_expr = SMCAtools.invert(
                self.ematrix,
                self.path_to('temp')
            )

            cc_sol = SMCAtools.solve_dep(
                CC_i_num,
                self.scaled_k0,
                self.scaled_l0,
                self.num_ind_fluxes,
                self.path_to('temp')
            )

            cc_sol, common_denom_expr = SMCAtools.fix_expressions(
                cc_sol,
                common_denom_expr,
                self.lmatrix,
                self.species_independent,
                self.species_dependent
            )

            cc_names = SMCAtools.build_cc_matrix(
                self.fluxes,
                self.fluxes_independent,
                self.species_independent,
                self.fluxes_dependent,
                self.species_dependent
            )

            cc_objects = SMCAtools.spawn_cc_objects(self.mod,
                                                    cc_names,
                                                    cc_sol,
                                                    common_denom_expr,
                                                    self._ltxe)

            self.cc_results = SMCAtools.make_CC_dot_dict(cc_objects)

            if internal_fixed:
                simpl_dic = SMCAtools.make_internals_dict(cc_sol,
                                                          cc_names,
                                                          common_denom_expr,
                                                          self.path_to('temp'))

                CC_block_counter = 0
                for each_common_denom_expr, name_num in simpl_dic.iteritems():
                    name_num[1], \
                        each_common_denom_expr = SMCAtools.fix_expressions(
                        name_num[1],
                        each_common_denom_expr,
                        self.lmatrix,
                        self.species_independent,
                        self.species_dependent
                    )

                    simpl_cc_objects = SMCAtools.spawn_cc_objects(self.mod,
                                                                  name_num[0],
                                                                  name_num[1],
                                                                  each_common_denom_expr,
                                                                  self._ltxe, )

                    CC_dot_dict = SMCAtools.make_CC_dot_dict(simpl_cc_objects)
                    setattr(self, 'cc_results_%s' %
                            CC_block_counter, CC_dot_dict)
                    CC_block_counter += 1

            self.CC_i_num = CC_i_num

        if auto_save_load:
            try:
                self.load_session()
            except:
                do_symca_internals(self)
                self.save_session()
        else:
            do_symca_internals(self)



import subprocess
from os import devnull
from os.path import join
# from os import mkdir
import sys
#from re import sub
from sympy import Symbol, sympify, nsimplify, fraction, S
from sympy.matrices import Matrix, diag, NonSquareMatrixError
from .ccobjects import CCBase, CCoef
from ...utils.misc import DotDict
from ...utils.misc import formatter_factory
from ...utils import ConfigReader
from ...utils.misc import ec_list, flux_list, ss_species_list


## Everything in this file can be a function rather than a static method
## better yet, almost everything can be part of symca. Finally everything can
## be in a single file.

all = ['SymcaToolBox']


class SymcaToolBox(object):
    """The class with the functions used to populate SymcaData. The project is
    structured in this way to abstract the 'work' needed to build the various
    matrices away from the SymcaData class."""

    @staticmethod
    def get_nmatrix(mod):
        """
        Returns a sympy matrix made from the N matrix in a Pysces model where
        the elements are in the same order as they appear in the k and l
        matrices in pysces.

        We need this to make calculations easier later on.
        """
        nmatrix = mod.nmatrix
        # swap columns around to same order as kmatrix, store in new matrix
        nmatrix_cols = nmatrix[:, mod.kmatrix_row]
        # swap rows around to same oder as lmatrix, store in a new matrix
        nmatrix_cols_rows = nmatrix_cols[mod.lmatrix_row, :]
        # create Sympy symbolic matrix from the numpy ndarray
        nmat = Matrix(nmatrix_cols_rows)
        return nmat

    @staticmethod
    def get_num_ind_species(mod):
        inds = len(mod.lmatrix_col)
        return inds

    @staticmethod
    def get_num_ind_fluxes(mod):
        inds = len(mod.kmatrix_col)
        return inds

    @staticmethod
    def get_species_vector(mod):
        """
        Returns a vector (sympy matrix) with the species in the correct order
        """
        slist = []
        # gets the order of the species from the lmatrix rows
        for index in mod.lmatrix_row:
            slist.append(mod.species[index])

        svector = Matrix(sympify(slist))
        #inds = len(mod.lmatrix_col)
        #Sind = Matrix(svector[:inds])
        #Sdep = Matrix(svector[inds:])
        return svector

    @staticmethod
    def get_fluxes_vector(mod):
        """
        Gets the dependent and independent fluxes (in the correct order)
        """

        jlist = []
        # gets the order of the fluxes from the kmatrix rows
        for index in mod.kmatrix_row:
            jlist.append('J_' + mod.reactions[index])
        jvector = Matrix(sympify(jlist))
        #inds = len(mod.kmatrix_col)
        #Jind = Matrix(jvector[:inds])
        #Jdep = Matrix(jvector[inds:])
        return jvector

    @staticmethod
    def substitute_fluxes(all_fluxes, kmatrix):
        """
        Substitutes equivalent fluxes in the kmatrix (e.i. dependent fluxes
        with independent fluxes or otherwise equal fluxes)
        """
        new_fluxes = all_fluxes[:, :]
        for row in xrange(kmatrix.rows - 1, -1, -1):
            for row_above in xrange(row - 1, -1, -1):
                if kmatrix[row, :] == kmatrix[row_above, :]:
                    new_fluxes[row] = new_fluxes[row_above]
        return new_fluxes

    @staticmethod
    def scale_matrix(all_elements, mat, inds):
        """
        Scales the k or l matrix.

        The procedure is the same for each matrix:
           (D^x)^(-1)   *          y         *        D^(x_i)

        Inverse diagonal   The matrix to be      The diagonal of
        of the x where     scaled. i.e. the      the independent x
        x is either the    k or l matrix         where x is the
        species or the                           species or the
        fluxes                                   fluxes

        """
        d_all_inv = diag(*all_elements).inv()
        d_inds = diag(*inds)
        scaled_matrix = d_all_inv * mat * d_inds
        return scaled_matrix

    @staticmethod
    def get_es_matrix(mod, nmatrix, fluxes, species):
        """
        Gets the esmatrix.

        Goes down the columns of the nmatrix (which holds the fluxes)
        to get the rows of the esmatrix.

        Nested loop goes down the rows of the nmatrix (which holds the species)
        to get the columns of the esmatrix

        so the format is

        ecReationN0_M0 ecReationN0_M1 ecReationN0_M2
        ecReationN1_M0 ecReationN1_M1 ecReationN1_M2
        ecReationN2_M0 ecReationN2_M1 ecReationN2_M2
        """
        nmat = nmatrix

        elas = []

        for col in range(nmat.cols):
            current_reaction = fluxes[col]
            elas_row = []
            for row in range(nmat.rows):
                current_species = species[row]
                ec_name = 'ec' + \
                          str(current_reaction)[2:] + '_' + str(
                    current_species)
                cond1 = getattr(mod, ec_name) != 0

                if cond1:
                    elas_row.append(ec_name)
                else:
                    elas_row.append(0)
            elas.append(elas_row)

        esmatrix = Matrix(elas)
        return esmatrix

    @staticmethod
    def get_es_matrix_no_mca(mod, nmatrix, fluxes, species):
        """
        Gets the esmatrix.

        Goes down the columns of the nmatrix (which holds the fluxes)
        to get the rows of the esmatrix.

        Nested loop goes down the rows of the nmatrix (which holds the species)
        to get the columns of the esmatrix

        so the format is

        ecReationN0_M0 ecReationN0_M1 ecReationN0_M2
        ecReationN1_M0 ecReationN1_M1 ecReationN1_M2
        ecReationN2_M0 ecReationN2_M1 ecReationN2_M2
        """
        nmat = nmatrix

        elas = []
        modifiers = dict(mod.__modifiers__)
        for col in range(nmat.cols):
            current_reaction = fluxes[col]
            elas_row = []
            for row in range(nmat.rows):
                current_species = species[row]
                ec_name = 'ec' + \
                          str(current_reaction)[2:] + '_' + str(
                    current_species)
                cond1 = nmat[row,col] != 0
                cond2 = str(current_species) in modifiers[str(current_reaction)[2:]]
                if cond1 or cond2:
                    elas_row.append(ec_name)
                else:
                    elas_row.append(0)
            elas.append(elas_row)

        esmatrix = Matrix(elas)
        return esmatrix

    @staticmethod
    def simplify_matrix(matrix):
        """
        Replaces floats with ints and puts elements with fractions
        on a single demoninator.
        """
        m = matrix[:, :]
        for i, e in enumerate(m):
            m[i] = nsimplify(e, rational=True).cancel()
        return m

    @staticmethod
    def adjugate_matrix(matrix):
        """
        Returns the adjugate matrix which is the transpose of the
        cofactor matrix.

        Contains code adapted from sympy.
        Specifically:

        cofactorMatrix()
        minorEntry()
        minorMatrix()
        cofactor()
        """

        def cofactor_matrix(mat):
            out = Matrix(mat.rows, mat.cols, lambda i, j:
            cofactor(mat, i, j))
            return out

        def minor_entry(mat, i, j):
            if not 0 <= i < mat.rows or not 0 <= j < mat.cols:
                raise ValueError(
                    "`i` and `j` must satisfy 0 <= i < `mat.rows` " +
                    "(%d)" % mat.rows + "and 0 <= j < `mat.cols` (%d)." % mat.cols)
            return SymcaToolBox.det_bareis(minor_matrix(mat, i, j))

        def minor_matrix(mat, i, j):
            if not 0 <= i < mat.rows or not 0 <= j < mat.cols:
                raise ValueError(
                    "`i` and `j` must satisfy 0 <= i < `mat.rows` " +
                    "(%d)" % mat.rows + "and 0 <= j < `mat.cols` (%d)." % mat.cols)
            m = mat.as_mutable()
            m.row_del(i)
            m.col_del(j)
            return m[:, :]

        def cofactor(mat, i, j):
            if (i + j) % 2 == 0:
                return minor_entry(mat, i, j)
            else:
                return -1 * minor_entry(mat, i, j)

        return cofactor_matrix(matrix).transpose()

    @staticmethod
    def det_bareis(matrix):
        """
        Adapted from original det_bareis function in Sympy 0.7.3.
        cancel() and expand() are removed from function to speed
        up calculations. Maxima will be used to simplify the result

        Original docstring below:

        Compute matrix determinant using Bareis' fraction-free
        algorithm which is an extension of the well known Gaussian
        elimination method. This approach is best suited for dense
        symbolic matrices and will result in a determinant with
        minimal number of fractions. It means that less term
        rewriting is needed on resulting formulae.
        """
        mat = matrix
        if not mat.is_square:
            raise NonSquareMatrixError()

        m, n = mat[:, :], mat.rows

        if n == 1:
            det = m[0, 0]
        elif n == 2:
            det = m[0, 0] * m[1, 1] - m[0, 1] * m[1, 0]
        else:
            sign = 1  # track current sign in case of column swap

            for k in range(n - 1):
                # look for a pivot in the current column
                # and assume det == 0 if none is found
                if m[k, k] == 0:
                    for i in range(k + 1, n):
                        if m[i, k] != 0:
                            m.row_swap(i, k)
                            sign *= -1
                            break
                    else:
                        return S.Zero

                # proceed with Bareis' fraction-free (FF)
                # form of Gaussian elimination algorithm
                for i in range(k + 1, n):
                    for j in range(k + 1, n):
                        d = m[k, k] * m[i, j] - m[i, k] * m[k, j]

                        if k > 0:
                            d /= m[k - 1, k - 1]

                        m[i, j] = d

            det = sign * m[n - 1, n - 1]

        return det

    @staticmethod
    def invert(matrix, path_to):
        """
        Returns the numerators of the inverted martix separately from the
        common denominator (the determinant of the matrix)
        """
        common_denom = SymcaToolBox.det_bareis(matrix)
        adjugate = SymcaToolBox.adjugate_matrix(matrix)

        common_denom = SymcaToolBox.maxima_factor(common_denom, path_to)
        #adjugate     = self._maxima_factor('/home/carl/test.txt',adjugate)

        cc_i_sol = adjugate, common_denom
        return cc_i_sol

    @staticmethod
    def maxima_factor(expression, path_to):
        """
        This function is equivalent to the sympy.cancel()
        function but uses maxima instead
        """

        maxima_in_file = join(path_to,'in.txt').replace('\\','\\\\')
        maxima_out_file = join(path_to,'out.txt').replace('\\','\\\\')
        if expression.is_Matrix:
            expr_mat = expression[:, :]
            # print expr_mat
            print 'Simplifying matrix with ' + str(len(expr_mat)) + ' elements'
            for i, e in enumerate(expr_mat):

                sys.stdout.write('*')
                sys.stdout.flush()
                if (i + 1) % 50 == 0:
                    sys.stdout.write(' ' + str(i + 1) + '\n')
                    sys.stdout.flush()
                # print e
                expr_mat[i] = SymcaToolBox.maxima_factor(e, path_to)
            sys.stdout.write('\n')
            sys.stdout.flush()
            return expr_mat
        else:
            batch_string = (
                'stardisp:true;stringout("'
                + maxima_out_file + '",factor(' + str(expression) + '));')
            # print batch_string
            with open(maxima_in_file, 'w') as f:
                f.write(batch_string)

            config = ConfigReader.get_config()
            if config['platform'] == 'win32':
                maxima_command = [config['maxima_path'], '--batch=' + maxima_in_file]
            else:
                maxima_command = ['maxima', '--batch=' + maxima_in_file]

            dn = open(devnull, 'w')
            subprocess.call(maxima_command, stdin=dn, stdout=dn, stderr=dn)
            simplified_expression = ''

            with open(maxima_out_file) as f:
                for line in f:
                    if line != '\n':
                        simplified_expression = line[:-2]
            frac = fraction(sympify(simplified_expression))
            # print frac[0].expand()/frac[1].expand()
            return frac[0].expand() / frac[1].expand()

    @staticmethod
    def solve_dep(cc_i_num, scaledk0, scaledl0, num_ind_fluxes, path_to):
        """
        Calculates the dependent control matrices from the independent control
        matrix CC_i_solution
        """

        j_cci_sol = cc_i_num[:num_ind_fluxes, :]
        s_cci_sol = cc_i_num[num_ind_fluxes:, :]

        j_ccd_sol = scaledk0 * j_cci_sol
        s_ccd_sol = scaledl0 * s_cci_sol

        tempmatrix = j_cci_sol
        for matrix in [j_ccd_sol, s_cci_sol, s_ccd_sol]:
            if len(matrix) != 0:
                tempmatrix = tempmatrix.col_join(matrix)

        cc_sol = tempmatrix

        cc_sol = SymcaToolBox.maxima_factor(cc_sol, path_to)

        # print len(j_cci_sol)
        # print len(j_ccd_sol)
        # print len(s_cci_sol)
        # print len(s_ccd_sol)

        return cc_sol

    @staticmethod
    def build_cc_matrix(j, jind, sind, jdep, sdep):
        """
        Produces the matrices j_cci, j_ccd, s_cci and s_ccd
        which holds the symbols for the independent and dependent flux control
        coefficients and the independent and dependent species control
        coefficients respectively
        """
        j_cci = []
        j_ccd = []
        s_cci = []
        s_ccd = []

        for Ji in jind:
            row = []
            for R in j:
                row.append('ccJ' + str(Ji)[2:] + '_' + str(R)[2:])
            j_cci.append(row)

        for Si in sind:
            row = []
            for R in j:
                row.append('cc' + str(Si) + '_' + str(R)[2:])
            s_cci.append(row)

        for Jd in jdep:
            row = []
            for R in j:
                row.append('ccJ' + str(Jd)[2:] + '_' + str(R)[2:])
            j_ccd.append(row)

        for Sd in sdep:
            row = []
            for R in j:
                row.append('cc' + str(Sd) + '_' + str(R)[2:])
            s_ccd.append(row)

        j_cci = Matrix(j_cci)
        j_ccd = Matrix(j_ccd)
        s_cci = Matrix(s_cci)
        s_ccd = Matrix(s_ccd)

        #cc_i = j_cci.col_join(s_cci)
        tempmatrix = j_cci
        for matrix in [j_ccd, s_cci, s_ccd]:
            if len(matrix) != 0:
                tempmatrix = tempmatrix.col_join(matrix)
        cc = tempmatrix

        # print len(j_cci)
        # print len(j_ccd)
        # print len(s_cci)
        # print len(s_ccd)

        return cc

    @staticmethod
    def get_fix_denom(lmatrix, species_independent, species_dependent):
        num_inds = len(species_independent)
        num_deps = len(species_dependent)
        if num_deps == 0:
            return sympify('1')
        else:
            dependent_ls = lmatrix[num_inds:, :]
            denom = sympify('1')
            for row in range(dependent_ls.rows):
                for each in dependent_ls[row, :] * species_independent * -1:
                    symbol_atoms = each.atoms(Symbol)
                    for symbol_atom in symbol_atoms:
                        if symbol_atom not in denom.atoms(Symbol):
                            denom = denom * symbol_atom
                            #denom = denom * each.atoms(Symbol).pop()
                denom = denom * species_dependent[row]
            return denom.nsimplify()

    def get_fix_denom_jannie(lmatrix, species_independent, species_dependent):
        num_inds = len(species_independent)
        num_deps = len(species_dependent)
        if num_deps == 0:
            return sympify('1')
        else:
            dependent_ls = lmatrix[num_inds:, :]
            denom = sympify('1')
            for row in range(dependent_ls.rows):
                den_new = sympify('1')
                for each in dependent_ls[row, :] * species_independent * -1:
                    symbol_atoms = each.atoms(Symbol)
                    for symbol_atom in symbol_atoms:
                        if den_new == 1:
                            den_new = den_new * symbol_atom
                        else:
                            den_new = den_new + symbol_atom
                            #denom = denom * each.atoms(Symbol).pop()
                if den_new == 1:
                    den_new = den_new * species_dependent[row]
                else:
                    den_new = den_new + species_dependent[row]
                denom = denom * den_new
            return denom.nsimplify()

    @staticmethod
    def fix_expressions(cc_num, common_denom_expr, lmatrix,
                        species_independent, species_dependent):

        fix_denom = SymcaToolBox.get_fix_denom(
            lmatrix,
            species_independent,
            species_dependent
        )
        fix = False
        cd_num, cd_denom = fraction(common_denom_expr)
        ret2 = cd_num


        if type(cc_num) is list:
            new_cc_num = cc_num[:]
        else:
            new_cc_num = cc_num[:, :]

        for i, each in enumerate(new_cc_num):
             new_cc_num[i] = ((each * cd_denom)).expand()

        for each in new_cc_num:
            for symb in fix_denom.atoms(Symbol):
                if symb in each.atoms(Symbol):
                    fix = True
                    break
            if fix: break

        if fix:
            for i, each in enumerate(new_cc_num):
                new_cc_num[i] = (each / fix_denom).expand()

            ret2 = (cd_num / fix_denom).expand()

        return new_cc_num, ret2

    @staticmethod
    def spawn_cc_objects(mod, cc_names, cc_sol, common_denom_exp, ltxe):


        common_denom_object = CCBase(mod,
                                     'common_denominator',
                                     common_denom_exp,
                                     ltxe)
        cc_object_list = [common_denom_object]

        for name, num in zip(cc_names, cc_sol):
            ccoef_object = CCoef(mod,
                                 str(name),
                                 num,
                                 common_denom_object,
                                 ltxe)

            cc_object_list.append(ccoef_object)
        return cc_object_list

    @staticmethod
    def make_internals_dict(cc_sol, cc_names, common_denom_expr, path_to):
        simpl_dic = {}
        for i, each in enumerate(cc_sol):
            expr = each / common_denom_expr
            expr = SymcaToolBox.maxima_factor(expr, path_to)
            num, denom = fraction(expr)
            if not simpl_dic.has_key(denom):
                simpl_dic[denom] = [[], []]
            simpl_dic[denom][0].append(cc_names[i])
            simpl_dic[denom][1].append(num)

        return simpl_dic

    @staticmethod
    def make_CC_dot_dict(cc_objects):
        CC = DotDict()
        for cc in cc_objects:
            CC[cc.name] = cc
        CC._make_repr('"$" + v.latex_name + "$"', 'v.value',
                      formatter_factory())
        return CC


    @staticmethod
    def build_inner_dict(cc_object):
        deep_dict = {}
        for key, value in cc_object.iteritems():
            if key != 'common_denominator':
                deepest_dict = {str(key): str(value.numerator)}
                deep_dict.update(deepest_dict)
        inner_dict = {str(cc_object.common_denominator.expression): deep_dict}
        return inner_dict

    @staticmethod
    def build_outer_dict(symca_object):
        containers = {}
        containers['cc_results'] = SymcaToolBox.build_inner_dict(
            getattr(symca_object, 'cc_results'))
        counter = 0
        while True:
            CC_obj_name = 'cc_results_{0}'.format(counter)
            try:
                CC_obj_dict = getattr(symca_object, CC_obj_name)
            except AttributeError:
                break
            containers[CC_obj_name] = SymcaToolBox.build_inner_dict(
                CC_obj_dict)
            counter += 1
        return containers

    @staticmethod
    def make_inner_dict(cc_container, cc_container_name):
        CC_dict = {}
        CC_dict[cc_container_name] = dict(zip(
            [cc.name for cc in cc_container.values() if
             cc.name is not 'common_denominator'],
            [cc.numerator for cc in cc_container.values() if
             cc.name is not 'common_denominator']))
        CC_dict[cc_container_name]['common_denominator'] = cc_container.common_denominator.expression
        return CC_dict


    @staticmethod
    def generic_populate(mod, function, value = 1):
        names = function(mod)
        for name in names:
            setattr(mod, name, value)

    @staticmethod
    def populate_with_fake_elasticities(mod):
        SymcaToolBox.generic_populate(mod, ec_list)

    @staticmethod
    def populate_with_fake_fluxes(mod):
        SymcaToolBox.generic_populate(mod, flux_list)

    @staticmethod
    def populate_with_fake_ss_concentrations(mod):
        SymcaToolBox.generic_populate(mod, ss_species_list)




        # OLD SAVE FUNCTIONS> Not as good as new ones
        # @staticmethod
        # def save_session(cc_list, common_denominator, path_to_pickle):
        #     mod = common_denominator.mod
        #     common_denominator.mod = ''
        #     for cc in cc_list:
        #         cc.mod = ''
        #         for cp in cc.control_patterns:
        #             cp.mod = ''
        #
        #     cc_list.append(common_denominator)
        #     with open(path_to_pickle, 'w') as f:
        #         pickle.dump(cc_list, f)
        #
        #     cc_list.pop()
        #     common_denominator.mod = mod
        #     for cc in cc_list:
        #         cc.mod = mod
        #         for cp in cc.control_patterns:
        #             cp.mod = mod
        #
        # @staticmethod
        # def load_session(mod, path_to_pickle):
        #     with open(path_to_pickle) as f:
        #         cc_list = pickle.load_session(f)
        #
        #     common_denominator = cc_list.pop()
        #
        #     common_denominator.mod = mod
        #     for cc in cc_list:
        #         cc.mod = mod
        #         for cp in cc.control_patterns:
        #             cp.mod = mod
        #
        #     cc_list.insert(0, common_denominator)
        #     return cc_list



from sympy import latex, sympify, Symbol

__all__ = ['LatexExpr']


def min_(string):
    """
    Returns a string stripped of underscores.

    Parameters
    ----------
    string : str
        A string containing underscores

    Returns
    -------
    values: str
        A string without underscores
    """
    return string.replace('_', '')


class LatexExpr(object):

    """docstring for LatexExpr"""

    def __init__(self, mod):
        super(LatexExpr, self).__init__()
        self.mod = mod
        self._subs_dict = None
        self._prc_subs = None
        self._added_tk = False
        self._tk_subs = None
        self._term_types = None

    def add_term_types(self, term_types):
        self._term_types = term_types

    def _add_tk_subs(self):
        if not self._subs_dict:
            self.subs_dict
        mod = self.mod
        j_subs = {}
        pec_subs = {}
        for term in self._term_types:
            # Fluxes:
            for reaction in mod.reactions:
                j_subs['J_%s_%s' % (reaction, term)] = 'J__%s_%s' % (min_(term),
                                                                     min_(reaction))
            for top_reaction in mod.reactions:
                for var_par in mod.species + mod.parameters:
                    o_ec = 'pec%s_%s_%s' % (top_reaction,
                                            var_par,
                                            term)
                    n_ec = 'varepsilon__%s_%s__%s' % (min_(top_reaction),
                                                      min_(var_par),
                                                      min_(term))
                    pec_subs[o_ec] = n_ec

        self._subs_dict.update(j_subs)
        self._subs_dict.update(pec_subs)
        self._added_tk = True

    @property
    def subs_dict(self):
        if not self._subs_dict:
            ec_subs = {}
            rc_subs = {}
            cc_subs = {}
            prc_subs = {}
            j_subs = {}
            sp_subs = {}
            par_subs = {}
            mod = self.mod

            # For clarity:
            # lots of code duplication
            # min_() strips string of underscores

            # Species:
            for species in mod.species:
                sp_subs[species] = min_(species)
                sp_subs[species + '_ss'] = min_(species) + '_ss'

            # Fluxes:
            for reaction in mod.reactions:
                j_subs[reaction] = min_(reaction)
                j_subs['J_' + reaction] = 'J_' + min_(reaction)

            # Parameters
            for par in mod.parameters:
                par_subs[par] = min_(par)

            # Control Coefficients:
            for base_reaction in mod.reactions:
                for top_species in mod.species:
                    o_cc = 'cc%s_%s' % (top_species, base_reaction)
                    n_cc = 'C__%s_%s' % (min_(top_species),
                                         min_(base_reaction))
                    cc_subs[o_cc] = n_cc

                for top_reaction in mod.reactions:
                    o_cc = 'ccJ%s_%s' % (top_reaction, base_reaction)
                    n_cc = 'C__J%s_%s' % (min_(top_reaction),
                                          min_(base_reaction))
                    cc_subs[o_cc] = n_cc

            # Elasticity Coefficients:
            for top_reaction in mod.reactions:
                for base_species in mod.species:
                    o_ec = 'ec%s_%s' % (top_reaction, base_species)
                    n_ec = 'varepsilon__%s_%s' % (min_(top_reaction),
                                                  min_(base_species))
                    ec_subs[o_ec] = n_ec

                for base_param in mod.parameters:
                    o_ec = 'ec%s_%s' % (top_reaction, base_param)
                    n_ec = 'varepsilon__%s_%s' % (min_(top_reaction),
                                                  min_(base_param))
                    ec_subs[o_ec] = n_ec

            # Response Coefficients:
            # I'm adding species together with parameter
            # This increases the size of the dictionary
            # but it ensures that only one LatexExpr object
            # is needed for something like RateChar (or SymCa)
            for base_param in set(mod.parameters + mod.species):
                for top_species in mod.species:
                    o_rc = 'rc%s_%s' % (top_species, base_param)
                    n_rc = 'R__%s_%s' % (min_(top_species),
                                         min_(base_param))
                    rc_subs[o_rc] = n_rc

                for top_reaction in mod.reactions:
                    o_rc = 'rcJ%s_%s' % (top_reaction, base_param)
                    n_rc = 'R__J%s_%s' % (min_(top_reaction),
                                          min_(base_param))
                    rc_subs[o_rc] = n_rc

            # Partial Response Coefficients:
            # I'm adding species together with parameter
            # This increases the size of the dictionary
            # but it ensures that only one LatexExpr object
            # is needed for something like RateChar (or SymCa)
            for base_param in set(mod.parameters + mod.species):
                for back_reaction in mod.reactions:
                    for top_species in mod.species:
                        o_prc = 'prc%s_%s_%s' % (top_species,
                                                 base_param,
                                                 back_reaction)
                        n_prc = 'R__%s_%s__%s' % (min_(top_species),
                                                  min_(base_param),
                                                  min_(back_reaction))
                        prc_subs[o_prc] = n_prc

                    for top_reaction in mod.reactions:
                        o_prc = 'prcJ%s_%s_%s' % (top_reaction,
                                                  base_param,
                                                  back_reaction)
                        n_prc = 'R__J%s_%s__%s' % (min_(top_reaction),
                                                   min_(base_param),
                                                   min_(back_reaction))
                        prc_subs[o_prc] = n_prc

            subs_dict = {}
            subs_dict.update(ec_subs)
            subs_dict.update(cc_subs)
            subs_dict.update(rc_subs)
            subs_dict.update(j_subs)
            subs_dict.update(sp_subs)
            subs_dict.update(prc_subs)
            subs_dict.update(par_subs)
            self._subs_dict = subs_dict

        return self._subs_dict

    @property
    def prc_subs(self):
        if not self._prc_subs:
            mod = self.mod
            prc_subs = {}
            for base_param in set(mod.parameters + mod.species):
                for back_reaction in mod.reactions:
                    for top_species in mod.species:
                        o_prc = 'R^{%s %s}_{%s}' % (min_(top_species),
                                                    min_(back_reaction),
                                                    min_(base_param))

                        n_prc = '\,^{%s}R^{%s}_{%s}' % (min_(back_reaction),
                                                        min_(top_species),
                                                        min_(base_param))

                        prc_subs[o_prc] = n_prc

                    for top_reaction in mod.reactions:
                        o_prc = 'R^{J%s %s}_{%s}' % (min_(top_reaction),
                                                     min_(back_reaction),
                                                     min_(base_param))

                        n_prc = '\,^{%s}R^{J%s}_{%s}' % (min_(back_reaction),
                                                         min_(top_reaction),
                                                         min_(base_param))

                        prc_subs[o_prc] = n_prc

            self._prc_subs = prc_subs

        return self._prc_subs

    @property
    def tk_subs(self):
        if not self._tk_subs:

            mod = self.mod
            tk_subs = {}
            pec_subs = {}
            j_subs = {}
            for term_type in self._term_types:
                for top_reaction in mod.reactions:
                    for base_param in mod.parameters + mod.species:
                        o_pec = '\\varepsilon^{%s %s}_{%s}' % (min_(top_reaction),
                                                               min_(term_type),
                                                               min_(base_param))

                        n_pec = '\\varepsilon^{{%s}_{%s}}_{%s}' % (min_(top_reaction),
                                                                   min_(
                                                                       term_type),
                                                                   min_(base_param))
                        pec_subs[o_pec] = n_pec

                    o_j = 'J^{%s}_{%s}' % (min_(term_type), min_(top_reaction))
                    n_j = 'J_{{%s}_{%s}}' % (
                        min_(top_reaction), min_(term_type))
                    j_subs[o_j] = n_j

            tk_subs.update(pec_subs)
            tk_subs.update(j_subs)
            self._tk_subs = tk_subs
        return self._tk_subs

    def expression_to_latex(self, expression,mul_symbol=None):
        if type(expression) == str:
            expression = sympify(expression)

        # symbol substitution in sympy takes longer for larger dicts
        # therefore I only get the symbols that I need
        # for mcanut model substitution of a 2 symbol expression
        # takes 3xxms for conversion with the full dict, but
        # only 1.9x - 2.2xms with the smaller_dict. Thats over 9000 times
        # faster!!! Ok only over 100X
        needed_symbols = expression.atoms(Symbol)
        smaller_dict = {}

        if self._term_types and not self._added_tk:
            self._add_tk_subs()

        for each in needed_symbols:
            each = str(each)
            if each[:2] == 'CP':
                smaller_dict[each] = each
            elif each in self.subs_dict:
                smaller_dict[each] = self.subs_dict[each]

        # using smaller_dict here instead of self.subs_dict
        latex_expr = latex(expression.subs(smaller_dict),
                           long_frac_ratio=10,
                           mul_symbol=mul_symbol)

        for k, v in self.prc_subs.iteritems():
            latex_expr = latex_expr.replace(k, v)

        if self._added_tk:
            for k, v in self.tk_subs.iteritems():
                latex_expr = latex_expr.replace(k, v)

        return latex_expr



from os import path
from collections import OrderedDict

from IPython.display import display, clear_output
from matplotlib import pyplot as plt
from matplotlib import transforms
from matplotlib import rcParams
from numpy import linspace
import ipywidgets as widgets
from pysces import ModelMap
from pysces import output_dir as psc_out_dir
import pysces
import gzip
import cPickle as pickle

from ..misc import *
from ...latextools import LatexExpr
from ... import modeltools

def save_data2d(data_2dobj, file_name):
    """
    Saves a Data2D object to a gzipped cPickle to a specified file name.
    """
    mod = data_2dobj.mod
    data_2dobj.mod = data_2dobj.mod.ModelFile
    with gzip.open(file_name, 'wb') as f:
        pickle.dump(data_2dobj, f)
    data_2dobj.mod = mod


def load_data2d(file_name, mod=None, ltxe=None):
    """
    Loads a gzipped cPickle file containing a Data2D object. Optionally
    a model can be provided (which is useful when loading data that
    reference the same model. For the same reason a LatexExpr object
    can be supplied.
    """
    with gzip.open(file_name, 'rb') as f:
        data_2dobj = pickle.load(f)
    if not mod:
        data_2dobj.mod = pysces.model(data_2dobj.mod)
    else:
        data_2dobj.mod = mod
    if ltxe:
        del data_2dobj._ltxe
        data_2dobj._ltxe = ltxe
    return data_2dobj

#matplotlib 1.5 breaks set_color_cycle functionality
#now we need cycler
from matplotlib import __version__ as mpl_version

use_cycler = False

from distutils.version import LooseVersion
if LooseVersion(mpl_version) >= LooseVersion('1.5.0'):
    from cycler import cycler
    use_cycler = True



exportLAWH = silence_print(pysces.write.exportLabelledArrayWithHeader)

"""
This whole module is fd in the a
"""
__all__ = ['LineData',
           'ScanFig',
           'Data2D',
           'load_data2d',
           'save_data2d',
           'SimpleData2D']


def _add_legend_viewlim(ax, **kwargs):
    """ Reset the legend in ax to only display lines that are
    currently visible in plot area """
    # THIS FUNCTION COMES FROM
    # http://matplotlib.1069221.n5.nabble.com/
    # Re-Limit-legend-to-visible-data-td18335.html
    label_objs = []
    label_texts = []
    # print "viewLim:", ax.viewLim
    for line in ax.lines:
        line_label = line.get_label()
        cond = line.get_visible() and \
               line_label and not line_label.startswith("_")
        if cond:
            line_bbox = transforms.Bbox.unit()
            line_bbox.update_from_data_xy(line.get_xydata())
            if ax.viewLim.overlaps(line_bbox):
                # print line_label, line_bbox
                label_objs.append(line)
                label_texts.append(line_label)
    if label_objs:
        return ax.legend(label_objs, label_texts, **kwargs)

    elif ax.get_legend():
        ax.get_legend().set_visible(False)
    else:
        ax.legend().set_visible(False)


class LineData(object):
    """
    An object that contains data and metadata used by ``ScanFig`` to draw a
    ``matplotlib`` line with interactivity.

    This object is used to initialise a ``ScanFig`` object together with a
    ``Data2D`` object. Once a ``ScanFig`` instance is initialised, the
    ``LineData`` objects are saved in a list ``_raw_line_data``. Changing
    any values there will have no effect on the output of the ``ScanFig``
    instance. Actual x,y data, ``matplotlib`` line metadata, and ``ScanFig``
    category metadata is stored.

    Parameters
    ----------
    name : str
        The name of the line. Will be used as a label if none is specified.
    x_data : array_like
        The x data.
    y_data : array_like
        The y data.
    categories : list, optional
        A list of categories that a line falls into. This will be used by
        ScanFig to draw buttons that enable/disable the line.
    properties : dict, optional
        A dictionary of properties of the line to be drawn. This dictionary
        will be used by the generic ``set()`` function of
        ``matplotlib.Lines.Line2D`` to set the properties of the line.

    See Also
    --------
    ScanFig
    Data2D
    RateChar
    """

    def __init__(self, name, x_data, y_data, categories=None, properties=None):
        super(LineData, self).__init__()
        self.name = name
        self.x = x_data
        self.y = y_data

        if categories:
            self.categories = categories
        else:
            self.categories = [self.name]
        if properties:
            self.properties = properties
        else:
            self.properties = {}

        self._update_attach_properties()

    def _update_attach_properties(self):
        """
        Attaches all properties in ``self.properties`` to the ``self``
        namespace.
        """
        # TODO Figure out why the properties are (or need to be) attached in this way. It seems unnecessary
        for k, v in self.properties.iteritems():
            setattr(self, k, v)

    def add_property(self, key, value):
        """
        Adds a property to the ``properties`` dictionary of the
        ``LineData`` object.

        The ``properties`` dictionary of ``LineData`` will be used by the
        generic ``set()`` function of ``matplotlib.Lines.Line2D``
        to set the properties of the line.

        Parameters
        ----------
        key : str
            The name of the ``matplotlib.Lines.Line2D`` property to be set.
        value : sting, int, bool
            The value of the property to be set. The type depends on the
            property.
        """

        self.properties.update({key, value})
        self._update_attach_properties()

class SimpleData2D(object):
    def __init__(self, column_names, data_array, mod=None):
        super(SimpleData2D, self).__init__()
        self.mod = mod
        if self.mod:
            self._ltxe = LatexExpr(mod)
        else:
            self._ltxe = None

        self.scan_results = DotDict()
        self.scan_results['scan_in'] = column_names[0]
        self.scan_results['scan_out'] = column_names[1:]
        self.scan_results['scan_range'] = data_array[:, 0]
        self.scan_results['scan_results'] = data_array[:, 1:]
        self.scan_results['scan_points'] = len(self.scan_results.scan_range)

        self._column_names = column_names
        self._scan_results = data_array
        self._setup_lines()

    def _setup_lines(self):
        """
        Sets up ``LineData`` objects that will be used to populate ``ScanFig``
        objects created by the ``plot`` method of ``Data2D``. These objects
        are stored in a list: ``self._lines``

        ``ScanFig`` takes a list of ``LineData`` objects as an argument and
        this method sets up that list. The ``self._column_categories``
        dictionary is used here.
        """
        lines = []
        for i, each in enumerate(self.scan_results.scan_out):
            if self._ltxe:
                label = self._ltxe.expression_to_latex(each)
            else:
                label = each

            line = LineData(name=each,
                            x_data=self.scan_results.scan_range,
                            y_data=self.scan_results.scan_results[:, i],
                            categories=[each],
                            properties={'label': '$%s$' % (label),
                                        'linewidth': 1.6})
            lines.append(line)
        self._lines = lines

    def plot(self):
        """
        Creates a ``ScanFig`` object using the data stored in the current
        instance of ``Data2D``

        Returns
        -------
        ``ScanFig``
            A `ScanFig`` object that is used to visualise results.
        """
        base_name = 'scan_fig'
        scan_fig = ScanFig(self._lines,
                           base_name=base_name,
                           ax_properties={'xlabel':
                                          self.scan_results.scan_in})
        return scan_fig

    def save_results(self, file_name=None, separator=',',fmt='%f'):
        """
        Saves data stores in current instance of ``Data2D`` as a comma
        separated file.

        Parameters
        ----------
        file_name : str, Optional (Default : None)
            The file name, extension and path under which data should be saved.
            If None the name will default to scan_data.csv and will be saved
            either under the directory specified under the directory specified
            in ``folder``.
        separator : str, Optional (Default : ',')
            The symbol which should be used to separate values in the output
            file.
        format : str, Optional (Default : '%f')
            Format for the data.
        """
        file_name = modeltools.get_file_path(working_dir=None,
                                             internal_filename='scan_fig',
                                             fmt='csv',
                                             fixed=self.scan_results.scan_in,
                                             file_name=file_name)

        scan_results = self._scan_results
        column_names = self._column_names

        try:
            exportLAWH(scan_results,
                       names=None,
                       header=column_names,
                       fname=file_name,
                       sep=separator,
                       format=fmt)
        except IOError as e:
            print e.strerror


class Data2D(object):
    """
    An object that wraps results from a PySCeS parameter scan.

    Results from parameter scan of timecourse are used to initialise this
    object which in turn is used to create a ``ScanFig`` object. Here results
    can easily be accessed and saved to disk.

    The ``Data2D`` is also responsible for setting up a ``ScanFig`` object from
    analysis results and therefore contains optional parameters for setting
    up this object.

    Parameters
    ----------
    mod : PysMod
        The model for which the parameter scan was performed.
    column_names : list of str
        The names of each column in the data_array. Columns should be arranged
        with the input values (scan_in, time) in the first column and the
        output values (scan_out) in the columns that follow.
    data_array : ndarray
        An array containing results from a parameter scan or tome simulation.
        Arranged as described above.
    ltxe : LatexExpr, optional (Default : None)
        A LatexExpr object that is used to convert PySCeS compatible
        expressions to LaTeX math. If None is supplied a new LatexExpr object
        will be instantiated. Sharing a single instance saves memory.
    analysis_method : str, Optional (Default : None)
        A string that indicates the name of the analysis method used to
        generate the results that populate ``Data2D``. This will determine
        where results are saved by ``Data2D`` as well as any ``ScanFig``
        objects that are produced by it.
    ax_properties : dict, Optional (Default : None)
        A dictionary of properties that will be used by ``ScanFig`` to adjust
        the appearance of plots. These properties should compatible with
        ``matplotlib.axes.AxesSubplot'' object in a way that its ``set``
        method can be used to change its properties. If none, a default
        ``ScanFig`` object is produced by the ``plot`` method.
    file_name : str, Optional (Default : None)
        The name that should be prepended to files produced any ``ScanFig``
        objects produced by ``Data2D``. If None, defaults to 'scan_fig'.
    additional_cat_classes : dict, Optional (Default : None)
        A dictionary containing additional line class categories for
        ``ScanFig`` construction. Each ``data_array`` column contains results
        representing a specific category of result (elasticity, flux,
        concentration) which in turn fall into a larger class of data types
        (All Coefficients). This dictionary defines which line classes fall
        into which class category. (k = category class; v = line categories)
    additional_cats : dict, Optional (Default : None)
        A dictionary that defines additional result categories as well as the
        lines that fall into these categories. (k = line category, v =
        lines in category).
    num_of_groups : int, Optional (Default : None)
        A number that defines the number of groups of lines. Used to ensure
        that the lines that are closely related (e.g. elasticities for one
        reaction) have colors assigned to them that are easily differentiable.
    working_dir : str, Optional (Default : None)
        This string sets the working directory directly and if provided
        supersedes ``analysis_method``.

    See Also
    --------
    ScanFig
    Data2D
    RateChar
    """

    def __init__(self,
                 mod,
                 column_names,
                 data_array,
                 ltxe=None,
                 analysis_method=None,
                 ax_properties=None,
                 file_name=None,
                 additional_cat_classes=None,
                 additional_cats=None,
                 num_of_groups=None,
                 working_dir=None,
                 category_manifest=None,
                 axvline=True):
        self.scan_results = DotDict()
        self.scan_results['scan_in'] = column_names[0]
        self.scan_results['scan_out'] = column_names[1:]
        self.scan_results['scan_range'] = data_array[:, 0]
        self.scan_results['scan_results'] = data_array[:, 1:]
        self.scan_results['scan_points'] = len(self.scan_results.scan_range)

        self._column_names = column_names
        self._scan_results = data_array

        if not category_manifest:
            category_manifest = {}
        self._category_manifest = category_manifest

        self.mod = mod

        scan_in = self.scan_results.scan_in

        if not analysis_method:
            if scan_in.lower() == 'time':
                analysis_method = 'simulation'
            elif hasattr(self.mod, scan_in):
                analysis_method = 'parameter_scan'
            else:
                analysis_method = 'custom'
        self._analysis_method = analysis_method

        if scan_in.lower() != 'time':
            try:
                self.mod.doMcaRC()
            except:
                pass

        if axvline:
            self._vline_val = None
            if scan_in.lower() != 'time' and hasattr(self.mod, scan_in):
                self._vline_val = getattr(self.mod, scan_in)

        if not ltxe:
            ltxe = LatexExpr(mod)
        self._ltxe = ltxe

        #TODO check if this is even needed

        self._fname_specified = False

        if not file_name:
            self._fname = 'scan_data'
        else:
            self._fname = file_name
            self._fname_specified = True


        #This is here specifically for the do_mca_scan method of pysces. If
        if not working_dir:
            working_dir = modeltools.make_path(mod=self.mod,
                                               analysis_method=self._analysis_method)
        self._working_dir = working_dir

        self._ax_properties_ = ax_properties

        # So in order for ScanFig to have all those nice buttons that are
        # organised so well we need to set it up beforehand. Basically
        # each different line has different categories of lines that it falls
        # into. Then each each of these categories falls into a category class.
        # Each ``_category_classes`` key represents a category class and the
        # value is a list of categories that fall into a class.
        #
        # The dictionary ``_scan_types`` contains the different categories that
        # a line can fall into (in addition to the category containing itself).
        # Here a keys is a category and value is a list of lines in this
        # category.
        #
        # Buttons will be arranged so that a category class is a label under
        # which all the buttons that toggle a certain category is arranged
        # under. For instance under the label'All Coefficients' will be the
        # buttons 'Elasticity Coefficients', 'Control Coefficients',
        # 'Response Coefficients etc.
        #
        # We also add _scan_types to the ``_category_classes`` so that each
        # individual line has its own button.
        # There will therefore be a button called 'Control Coefficients' that
        # fall under the 'All Coefficients' category class label as well as a
        # label for the category class called 'Control Coefficients' under
        # which all the  different control coefficient buttons will be
        # arranged.


        if not additional_cat_classes:
            additional_cat_classes = {}
        self._additional_cat_classes = additional_cat_classes

        if not additional_cats:
            additional_cats = {}
        self._additional_cats = additional_cats


        self._setup_lines()

        if num_of_groups:
            self._lines = group_sort(self._lines, num_of_groups)

    @property
    def _category_classes(self):
        category_classes = OrderedDict([('All Coefficients',
                          ['Elasticity Coefficients',
                           'Control Coefficients',
                           'Response Coefficients',
                           'Partial Response Coefficients',
                           'Control Patterns']),
                         ('All Fluxes/Reactions/Species/Parameters',
                          ['Flux Rates',
                           'Reaction Rates',
                           'Species Concentrations',
                           'Steady-State Species Concentrations',
                           'Parameters'])])

        additional_cat_classes = self._additional_cat_classes
        for k, v in additional_cat_classes.iteritems():
            if k in category_classes:
                lst = category_classes[k]
                new_lst = list(set(lst + v))
                category_classes[k] = new_lst
            else:
                category_classes[k] = v
        category_classes.update(self._scan_types)
        return category_classes

    @property
    def _scan_types(self):
        scan_types = OrderedDict([
        ('Flux Rates', ['J_' + reaction for reaction in self.mod.reactions]),
        ('Reaction Rates', [reaction for reaction in self.mod.reactions]),
        ('Species Concentrations', self.mod.species + self.mod.fixed_species),
        ('Steady-State Species Concentrations',[sp + '_ss' for sp in self.mod.species]),
        ('Elasticity Coefficients', ec_list(self.mod)),
        ('Control Coefficients', cc_list(self.mod)),
        ('Response Coefficients', rc_list(self.mod)),
        ('Partial Response Coefficients', prc_list(self.mod)),
        ('Control Patterns', ['CP{:3}'.format(n).replace(' ','0')
                              for n in range(1, len(self._column_names))]),
        ('Parameters', self.mod.parameters)])

        additional_cats = self._additional_cats
        if additional_cats:
            for k, v in additional_cats.iteritems():
                if k in scan_types:
                    lst = scan_types[k]
                    new_lst = list(set(lst + v))
                    scan_types[k] = new_lst
                else:
                    scan_types[k] = v
        return scan_types

    @property
    def _column_categories(self):
        """
        This method sets up the categories for each data column stored by this
        object. These categories are stored in a dictionary as
        ``self._column_categories``.

        Each line falls into its own category as well as another category
        depending on what type of data it represents. So 'Species1' will
        fall into the category 'Species1' as well as 'Species Concentrations'
        Therefore the ``ScanFig`` buttons labelled 'Species1' and 'Species
        Concentrations' need to be toggled on for the line representing
        the parameter scan results of Species1 to be visible on the
        ``ScanFig`` figure.
        """
        scan_types = self._scan_types
        column_categories = {}
        for column in self.scan_results.scan_out:
            column_categories[column] = [column]
            for k, v in scan_types.iteritems():
                if column in v:
                    column_categories[column].append(k)
                    break

        return column_categories

    def _setup_lines(self):
        """
        Sets up ``LineData`` objects that will be used to populate ``ScanFig``
        objects created by the ``plot`` method of ``Data2D``. These objects
        are stored in a list: ``self._lines``

        ``ScanFig`` takes a list of ``LineData`` objects as an argument and
        this method sets up that list. The ``self._column_categories``
        dictionary is used here.
        """
        _column_categories = self._column_categories
        lines = []
        for i, each in enumerate(self.scan_results.scan_out):
            line = LineData(name=each,
                            x_data=self.scan_results.scan_range,
                            y_data=self.scan_results.scan_results[:, i],
                            categories=_column_categories[each],
                            properties={'label':
                                            '$%s$' %
                                            (self._ltxe.expression_to_latex(
                                                each)),
                                        'linewidth': 1.6})
            lines.append(line)
        self._lines = lines

    @property
    def _ax_properties(self):
        """
        Internal property of ``Data2D``. If no ``ax_properties`` argument is
        specified in __init__ this property defines the xlabel of the
        ``ScanFig`` object depending on the value of ``self.scan_in``.

        """
        if not self._ax_properties_:
            self._ax_properties_ = {'xlabel': self._x_name}
        return self._ax_properties_

    @property
    def _x_name(self):
        mm = ModelMap(self.mod)
        species = mm.hasSpecies()
        x_name = ''
        # TODO Enable lower case "time" as well as well as making generic for minutes/hours
        if self.scan_results.scan_in.lower() == 'time':
            x_name = 'Time'
        elif self.scan_results.scan_in in species:
            x_name = '[%s]' % self.scan_results.scan_in
        elif self.scan_results.scan_in in self.mod.parameters:
            x_name = self.scan_results.scan_in
        return x_name

    def plot(self):
        """
        Creates a ``ScanFig`` object using the data stored in the current
        instance of ``Data2D``

        Returns
        -------
        ``ScanFig``
            A `ScanFig`` object that is used to visualise results.
        """
        if self._fname_specified:
            base_name = self._fname
        else:
            base_name = 'scan_fig'
        scan_fig = ScanFig(self._lines,
                           category_classes=self._category_classes,
                           ax_properties=self._ax_properties,
                           working_dir=path.join(self._working_dir,
                                                 self.scan_results.scan_in, ),
                           base_name=base_name, )

        for k,v in self._category_manifest.iteritems():
            scan_fig.toggle_category(k,v)

        if self._vline_val:
            scan_fig.ax.axvline(self._vline_val, ls=':', color='gray')

        return scan_fig

    def save_results(self, file_name=None, separator=',',fmt='%f'):
        """
        Saves data stores in current instance of ``Data2D`` as a comma
        separated file.

        Parameters
        ----------
        file_name : str, Optional (Default : None)
            The file name, extension and path under which data should be saved.
            If None the name will default to scan_data.csv and will be saved
            either under the directory specified under the directory specified
            in ``folder``.
        separator : str, Optional (Default : ',')
            The symbol which should be used to separate values in the output
            file.
        format : str, Optional (Default : '%f')
            Format for the data.
        """
        file_name = modeltools.get_file_path(working_dir=self._working_dir,
                                             internal_filename=self._fname,
                                             fmt='csv',
                                             fixed=self.scan_results.scan_in,
                                             file_name=file_name)

        scan_results = self._scan_results
        column_names = self._column_names

        try:
            exportLAWH(scan_results,
                       names=None,
                       header=column_names,
                       fname=file_name,
                       sep=separator,
                       format=fmt)
        except IOError as e:
            print e.strerror


class ScanFig(object):
    """
    Uses data in the form of a list of LineData objects to display interactive
    plots.

    Interactive plots can be customised in terms of which data is visible at
    any one time by simply clicking a button to toggle a line. Matplotlib
    figures are used internally, therefore ScanFig figures can be altered
    by changing the properties of the internal figure.

    Parameters
    ----------
    line_data_list : list of LineData objects
        A LineData object contains the information needed to draw a single
        curve on a matplotlib figure. Here a list of these objects are used
        to populate the internal matplotlib figure with the various curves
        that represent the results of a parameter scan or simulation.
    category_classes : dict, Optional (Default : None)
        Each line on a ScanFig plot falls into a different category. Each of
        these categories in turn fall into a different class. Each category
        represents a button which toggles the lines which fall into the
        category while the button is arranged under a label which is
        represented by a category class. Each key in this dict is a category
        class and the value is a list of categories that fall into this class.
        If None all categories will fall into the same class.
    fig_properties : dict, Optional (Default : None)
        A dictionary of properties that will be used to adjust the appearance
        of the figure. These properties should compatible with
        ``matplotlib.figure.Figure'' object in a way that its ``set``
        method can be used to change its properties. If None, default
        matplotlib figure properties will be used.
    ax_properties : dict, Optional (Default : None)
        A dictionary of properties that will be used to adjust the appearance
        of plot axes. These properties should compatible with
        ``matplotlib.axes.AxesSubplot'' object in a way that its ``set``
        method can be used to change its properties. If None default matplotlib
        axes properties will be used.
    base_name : str, Optional (Default : None)
        Base name that will be used when an image is saved by ``ScanFig``. If
        None, then ``scan_fig`` will be used.
    working_dir : str, Optional (Default : None)
        The directory in which files figures will be saved. If None, then it
        will default to the directory specified in ``pysces.output_dir``.


    See Also
    --------
    LineData
    Data2D
    """

    def __init__(self, line_data_list,
                 category_classes=None,
                 fig_properties=None,
                 ax_properties=None,
                 base_name=None,
                 working_dir=None):

        super(ScanFig, self).__init__()

        rcParams.update({'font.size': 16})

        self._categories_ = None
        self._categories_status = None
        self._lines_ = None
        self._widgets_ = None
        self._figure_widgets_ = None
        self._raw_line_data = line_data_list

        # figure setup
        plt.ioff()
        self.fig = plt.figure(figsize=(10, 5.72))
        if fig_properties:
            self.fig.set(**fig_properties)

        # axis setup
        self.ax = self.fig.add_subplot(111)
        if ax_properties:
            self.ax.set(**ax_properties)

        # colourmap_setup
        # at the moment this is very basic and could be expanded
        # it would be useful to set it up based on category somehow
        cmap = plt.get_cmap('Set1')(
            linspace(0, 1.0, len(line_data_list)))
        if use_cycler:
            col_cycler = cycler('color',cmap)
            self.ax.set_prop_cycle(col_cycler)
        else:
            self.ax.set_color_cycle(cmap)

        if category_classes:
            new_cat_classes = OrderedDict()
            for k, v in category_classes.iteritems():
                for each in self._categories.iterkeys():
                    if each in v:
                        if not k in new_cat_classes:
                            new_cat_classes[k] = []
                        new_cat_classes[k].append(each)
            self._category_classes = new_cat_classes
        else:
            self._category_classes = {'': [k for k in self._categories]}

        if base_name:
            self._base_name = base_name
        else:
            self._base_name = 'scan_fig'

        if working_dir:
            self._working_dir = working_dir
        else:
            self._working_dir = psc_out_dir

        self._save_counter = 0

        self._lines

        if 'backend_inline' in rcParams['backend']:
            plt.close()
        self._save_button_ = None


    @property
    def _save_button(self):
        if not self._save_button_:
            def save(clicked):
                self.save()

            self._save_button_ = widgets.Button()
            self._save_button_.description = 'Save'
            self._save_button_.on_click(save)
        return self._save_button_

    def show(self):
        """
        Displays the figure.

        Depending on the matplotlib backend this function will either display
        the figure inline if running in an ``IPython`` notebook with the
        ``--pylab=inline`` switch or with the %matplotlib inline IPython line
        magic, alternately it will display the figure as determined by the
        ``rcParams['backend']`` option of ``matplotlib``. Either the inline or
        nbagg backends are recommended.

        See Also
        --------
        interact
        adjust_figure
        """

        _add_legend_viewlim(
            self.ax,
            bbox_to_anchor=(0, -0.17),
            ncol=3,
            loc=2,
            borderaxespad=0.)

        if 'backend_inline' in rcParams['backend']:
            clear_output(wait=True)
            display(self.fig)
        else:
            self.fig.show()

    def save(self, file_name=None, dpi=None, fmt=None, include_legend=True):
        """
        Saves the figure in it's current configuration.

        Parameters
        ----------
        file_name : str, Optional (Default : None)
            The file name to be used. If None is provided the file will be saved
            to ``working_dir/base_name.fmt``
        dpi : int, Optional (Default : None)
            The dpi to use. Defaults to 180.
        fmt : str, Optional (Default : None)
            The image format to use. Defaults to ``svg``. If ``file_name``
            contains a valid extension it will supersede ``fmt``.

        """
        if not fmt:
            fmt = 'svg'

        if not dpi:
            dpi = 180

        file_name = modeltools.get_file_path(working_dir=self._working_dir,
                                             internal_filename=self._base_name,
                                             fmt=fmt,
                                             file_name=file_name)
        fmt = modeltools.get_fmt(file_name)
        if include_legend:
            self.fig.savefig(file_name,
                             format=fmt,
                             dpi=dpi,
                             bbox_extra_artists=(self.ax.get_legend(),),
                             bbox_inches='tight')
        else:
            leg = self.ax.legend_
            self.ax.legend_ = None
            self.fig.savefig(file_name,
                             format=fmt,
                             dpi=dpi,)
            self.ax.legend_ = leg

    @property
    def _widgets(self):
        if not self._widgets_:
            widget_classes = OrderedDict()
            for k in self._category_classes.iterkeys():
                box = widgets.HBox()
                box.layout.display = 'flex-flow'
                widget_classes[k] = box
            def oc(cat):
                def on_change(name, value):
                    self.toggle_category(cat, value)
                    self.show()

                return on_change


            width = self._find_button_width()

            for each in self._categories:
                w = widgets.ToggleButton()
                w.description = each
                w.width = width
                w.value = self.categories_status[each]
                on_change = oc(each)
                w.on_trait_change(on_change, 'value')
                for k, v in self._category_classes.iteritems():
                    if each in v:
                        widget_classes[k].children += (w),

            # this is needed to sort widgets according to alphabetical order
            for k, v in widget_classes.iteritems():
                children_list = list(v.children)
                names = [getattr(widg, 'description')
                         for widg in children_list]
                names.sort()

                new_children_list = []
                for name in names:
                    for child in children_list:
                        if child.description == name:
                            new_children_list.append(child)
                v.children = tuple(new_children_list)

            self._widgets_ = widget_classes
        return self._widgets_

    @property
    def _figure_widgets(self):
        """
        Instantiates the widgets that will be used to adjust the figure.

        At the moment widgets for manipulating the following paramers
        are available:

            minimum and maximum x values on the x axis
            minimum and maximum y values on the y axis
            the scale of the x and y axis i.e. log vs linear

        The following are possible TODOs:

            figure size
            y label
            x label
            figure title
        """

        def convert_scale(val):
            """
            Converts between str and bool for the strings 'log' and 'linear'

            The string 'log' returns True, while True returns 'log'.
            The string 'linear' returns False, while False returns 'linear'

            Parameters
            ----------
            val : str, bool
                The value to convert.

            Returns
            -------
            value : str, bool
                The conversion of the parameter ``val``

            Examples
            --------
            >>> convert_scale('log')
            True
            >>> convert_scale(False)
            'linear'
            """

            if type(val) == bool:
                if val is True:
                    return 'log'
                elif val is False:
                    return 'linear'
            else:
                if val == 'log':
                    return True
                elif val == 'linear':
                    return False

        def c_v(val):

            if val <= 0:
                return 0.001
            else:
                return val

        if not self._figure_widgets_:
            min_x = widgets.FloatText()
            max_x = widgets.FloatText()
            min_x.value, max_x.value = self.ax.get_xlim()
            min_x.description = 'min'
            max_x.description = 'max'

            min_y = widgets.FloatText()
            max_y = widgets.FloatText()
            min_y.value, max_y.value = self.ax.get_ylim()
            min_y.description = 'min'
            max_y.description = 'max'

            log_x = widgets.Checkbox()
            log_y = widgets.Checkbox()
            log_x.value = convert_scale(self.ax.get_xscale())
            log_y.value = convert_scale(self.ax.get_yscale())
            log_x.description = 'x_log'
            log_y.description = 'y_log'

            apply_btn = widgets.Button()
            apply_btn.description = 'Apply'

            def set_values(clicked):
                if log_x.value is True:
                    min_x.value = c_v(min_x.value)
                    max_x.value = c_v(max_x.value)

                self.ax.set_xlim([min_x.value, max_x.value])

                if log_y.value is True:
                    min_y.value = c_v(min_y.value)
                    max_y.value = c_v(max_y.value)

                self.ax.set_ylim([min_y.value, max_y.value])

                self.ax.set_xscale(convert_scale(log_x.value))
                self.ax.set_yscale(convert_scale(log_y.value))

                self.show()

            apply_btn.on_click(set_values)

            x_lims = widgets.HBox(children=[min_x, max_x])
            y_lims = widgets.HBox(children=[min_y, max_y])
            lin_log = widgets.HBox(children=[log_x, log_y])
            apply_con = widgets.HBox(children=[apply_btn])

            _figure_widgets_ = OrderedDict()
            _figure_widgets_['X axis limits'] = x_lims
            _figure_widgets_['Y axis limits'] = y_lims
            _figure_widgets_['Axis scale'] = lin_log
            _figure_widgets_['    '] = apply_con

            self._figure_widgets_ = _figure_widgets_
        return self._figure_widgets_

    @property
    def _categories(self):
        if not self._categories_:
            main_cats = []
            cats = []
            for each in self._raw_line_data:
                cats += each.categories
                main_cats.append(each.categories[0])
            cats = list(set(cats))

            cat_dict = {}
            for each in cats:
                cat_dict[each] = []

            for each in self._raw_line_data:
                line = self._lines[each.name]
                for cat in each.categories:
                    cat_dict[cat].append(line)

            self._categories_ = cat_dict
        return self._categories_

    @property
    def category_names(self):
        return self._categories.keys()

    @property
    def categories_status(self):
        if not self._categories_status:
            cat_stat_dict = {}
            for each in self._categories:
                cat_stat_dict[each] = False

            self._categories_status = cat_stat_dict
        return self._categories_status

    @property
    def _lines(self):
        if not self._lines_:
            lines = {}
            for i, each in enumerate(self._raw_line_data):
                line, = self.ax.plot(each.x, each.y)

                # set width to a default width of 2
                # bc the default value of one is too low
                line.set_linewidth(2)
                if each.properties:
                    line.set(**each.properties)
                else:
                    line.set_label(each.name)

                line.set_visible(False)

                lines[each.name] = line
            self._lines_ = lines
        return self._lines_

    @property
    def line_names(self):
        lines = self._lines.keys()
        lines.sort()
        return lines


    def toggle_line(self, name, value):
        """
        Changes the visibility of a certain line.

        When used a specific line's visibility is changed according to the
        ``value`` provided.

        Parameters
        ----------
        name: str
            The name of the line to change.
        value: bool
            The visibility status to change the line to (True for visible,
            False for invisible).

        See Also
        --------
        toggle_category


        """
        self._lines[name].set_visible(value)

    def toggle_category(self, cat, value):
        """
        Changes the visibility of all the lines in a certain line category.

        When used all lines in the provided category's  visibility is changed
        according to the ``value`` provided.

        Parameters
        ----------
        cat: str
            The name of the category to change.
        value: bool
            The visibility status to change the lines to (True for visible,
            False for invisible).

        See Also
        --------
        toggle_line

        """

        # get the visibility status of the category eg. True/False
        self.categories_status[cat] = value
        # get all the other categories
        other_cats = self._categories.keys()
        other_cats.pop(other_cats.index(cat))
        # self.categories is a dict with categories as keys
        # and list of lines that fall within a category
        # as a value. So for each line that falls in a cat
        for line in self._categories[cat]:
            # The visibility for a line has not changed at the start of
            # the loop
            in_other_cats = False
            # A line can also fall within another category
            other_cat_stats = []

            for each in other_cats:
                if line in self._categories[each]:
                    other_cat_stats.append(self.categories_status[each])
                    in_other_cats = True
            # If a line is never in any other categories
            # just set its  visibility as it is dictated by
            # its category status.
            if in_other_cats:
                visibility = all([value] + other_cat_stats)
                line.set_visible(visibility)
            else:
                line.set_visible(value)

    def interact(self):
        """
        Displays the figure in a IPython/Jupyter notebook together with buttons
        to toggle the visibility of certain lines.

        See Also
        --------
        show
        adjust_figure
        """
        self.show()
        for k, v in self._widgets.iteritems():
            if len(v.children) > 0:
                head = widgets.Label(value=k)
                display(head)
                display(v)
                v._css = [(None, 'flex-wrap', 'wrap'), ]
                # v.remove_class('vbox')
                # v.add_class('hbox')
                # v.set_css({'flex-wrap': 'wrap'})
        display(widgets.Label(value='$~$'))
        display(self._save_button)
        for boxes in self._widgets.itervalues():
            for button in boxes.children:
                button.value = self.categories_status[button.description]
                # self._save_button.remove_class('vbox')
                # self._save_button.add_class('hbox')

    def adjust_figure(self):
        """
        Provides widgets to set the limits and scale (log/linear) of the figure.

        As with ``interact``, the plot is displayed in the notebook. Here
        no widgets are provided the change the visibility of the data
        displayed on the plot, rather controls to set the limits and scale are
        provided.

        See Also
        --------
        show
        interact

        """
        self.show()
        for k, v in self._figure_widgets.iteritems():
            if len(v.children) > 0:
                head = widgets.Label(value=k)
                display(head)
                display(v)
                # v.remove_class('vbox')
                # v.add_class('hbox')
                v._css = [(None, 'flex-wrap', 'wrap'), ]
        display(widgets.Label(value='$~$'))
        display(self._save_button)
        # self._save_button.remove_class('vbox')
        # self._save_button.add_class('hbox')

    def _find_button_width(self):
        longest = sorted([len(each) for each in self._categories])[-1]
        if longest > 14:
            width_px = (longest - 14) * 5 + 145
            width = str(width_px) + 'px'
        else:
            width = '145px'
        return width



import time
## test
class EventfulDict(dict):
    """Eventful dictionary"""

    def __init__(self, *args, **kwargs):
        """Sleep is an optional float that allows you to tell the
        dictionary to hang for the given amount of seconds on each
        event.  This is usefull for animations."""
        self._sleep = kwargs.get('sleep', 0.0)
        self._add_callbacks = []
        self._del_callbacks = []
        self._set_callbacks = []
        dict.__init__(self, *args, **kwargs)

    def on_add(self, callback, remove=False):
        self._register_callback(self._add_callbacks, callback, remove)
    def on_del(self, callback, remove=False):
        self._register_callback(self._del_callbacks, callback, remove)
    def on_set(self, callback, remove=False):
        self._register_callback(self._set_callbacks, callback, remove)
    def _register_callback(self, callback_list, callback, remove=False):
        if callable(callback):
            if remove and callback in callback_list:
                callback_list.remove(callback)
            elif not remove and not callback in callback_list:
                callback_list.append(callback)
        else:
            raise Exception('Callback must be callable.')

    def _handle_add(self, key, value):
        self._try_callbacks(self._add_callbacks, key, value)
        self._try_sleep()
    def _handle_del(self, key):
        self._try_callbacks(self._del_callbacks, key)
        self._try_sleep()
    def _handle_set(self, key, value):
        self._try_callbacks(self._set_callbacks, key, value)
        self._try_sleep()
    def _try_callbacks(self, callback_list, *pargs, **kwargs):
        for callback in callback_list:
            callback(*pargs, **kwargs)

    def _try_sleep(self):
        if self._sleep > 0.0:
            time.sleep(self._sleep)

    def __setitem__(self, key, value):
        return_val = None
        exists = False
        if key in self:
            exists = True

        # If the user sets the property to a new dict, make the dict
        # eventful and listen to the changes of it ONLY if it is not
        # already eventful.  Any modification to this new dict will
        # fire a set event of the parent dict.
        if isinstance(value, dict) and not isinstance(value, EventfulDict):
            new_dict = EventfulDict(value)

            def handle_change(*pargs, **kwargs):
                self._try_callbacks(self._set_callbacks, key, dict.__getitem__(self, key))

            new_dict.on_add(handle_change)
            new_dict.on_del(handle_change)
            new_dict.on_set(handle_change)
            return_val = dict.__setitem__(self, key, new_dict)
        else:
            return_val = dict.__setitem__(self, key, value)

        if exists:
            self._handle_set(key, value)
        else:
            self._handle_add(key, value)
        return return_val

    def __delitem__(self, key):
        return_val = dict.__delitem__(self, key)
        self._handle_del(key)
        return return_val

    def pop(self, key):
        return_val = dict.pop(self, key)
        if key in self:
            self._handle_del(key)
        return return_val

    def popitem(self):
        popped = dict.popitem(self)
        if popped is not None and popped[0] is not None:
            self._handle_del(popped[0])
        return popped

    def update(self, other_dict):
        for (key, value) in other_dict.items():
            self[key] = value

    def clear(self):
        for key in list(self.keys()):
            del self[key]


import sys
from os import path, devnull


from numpy.ma import log10
from numpy import array, errstate, nanmin, nanmax, nonzero, float64,\
    bool_, string_, ones
from pysces.PyscesModel import PysMod
from IPython.display import HTML
from sympy import sympify
from functools import wraps
from ..config import ConfigReader

__all__ = ['cc_list',
           'ec_list',
           'rc_list',
           'prc_list',
           'silence_print',
           'DotDict',
           'PseudoDotDict',
           'is_number',
           'formatter_factory',
           'html_table',
           'do_safe_state',
           'find_min',
           'find_max',
           'split_coefficient',
           'ec_dict',
           'cc_dict',
           'rc_dict',
           'prc_dict',
           'group_sort',
           'extract_model',
           'get_value',
           'get_value_eval',
           'get_value_sympy',
           'print_f',
           'stringify',
           'is_iterable',
           'scanner_range_setup',
           'is_linear',
           'column_multiply',
           'unix_to_windows_path',
           'flux_list',
           'ss_species_list',
           'get_filename_from_caller',
           'memoize',
           'is_reaction',
           'is_species',
           'is_parameter',
           'is_variable',
           'is_attr',
           'is_mca_coef',
           'is_ec',
           'is_cc',
           'is_rc',
           'is_prc',]


def memoize(function):
    memo = {}
    @wraps(function)
    def wrapper(*args):
        if args in memo:
            return memo[args]
        else:
            return_val = function(*args)
            memo[args] = return_val
            return return_val
    return wrapper

@memoize
def is_species(attr, model):
    if attr.endswith('_ss'):
        attr = attr[:-3]
    if attr in model.species:
        return True
    else:
        return False

@memoize
def is_reaction(attr, model):
    if attr.endswith('J_'):
        attr = attr[2:]
    if attr in model.reactions:
        return True
    else:
        return False

@memoize
def is_parameter(attr, model):
    if attr in model.parameters:
        return True
    else:
        return False

@memoize
def is_variable(attr, model):
    if is_species(attr,model) or is_reaction(attr,model) or is_mca_coef(attr, model):
        return True
    else:
        return False

@memoize
def is_attr(attr, model):
    if is_variable(attr, model) or is_parameter(attr, model) or is_mca_coef(attr, model):
        return True
    else:
        return False

@memoize
def is_cc(attr, model):
    vals = split_coefficient(attr, model)
    if vals is not None and vals[0] == 'cc':
        return True
    else:
        return False

@memoize
def is_ec(attr, model):
    vals = split_coefficient(attr, model)
    if vals is not None and vals[0] == 'ec':
        return True
    else:
        return False

@memoize
def is_prc(attr, model):
    vals = split_coefficient(attr, model)
    if vals is not None and vals[0] == 'prc':
        return True
    else:
        return False

@memoize
def is_rc(attr, model):
    vals = split_coefficient(attr, model)
    if vals is not None and vals[0] == 'rc':
        return True
    else:
        return False
@memoize
def is_mca_coef(attr, model):
    mca_func_list = [is_rc, is_prc, is_ec, is_ec]
    return any([func(attr, model) for func in mca_func_list])

def get_filename_from_caller():
    try:
        caller = sys._getframe(1)
    except ValueError:
        globals = sys.__dict__
        lineno = 1
    else:
        globals = caller.f_globals
        lineno = caller.f_lineno
    if '__name__' in globals:
        module = globals['__name__']
    else:
        module = "<string>"
    filename = globals.get('__file__')
    if filename:
        fnl = filename.lower()
        if fnl.endswith((".pyc", ".pyo")):
            filename = filename[:-1]
    else:
        if module == "__main__":
            try:
                filename = sys.argv[0]
            except AttributeError:
                # embedded interpreters don't have sys.argv, see bug #839151
                filename = '__main__'
        if not filename:
            filename = module
    return filename


def unix_to_windows_path(path_to_convert, drive_letter='C'):
    """
    For a string representing a POSIX compatible path (usually
    starting with either '~' or '/'), returns a string representing an
    equivalent Windows compatible path together with a drive letter.

    Parameters
    ----------
    path_to_convert : string
        A string representing a POSIX path
    drive_letter : string (Default : 'C')
        A single character string representing the desired drive letter

    Returns
    -------
    string
        A string representing a Windows compatible path.
    """
    if path_to_convert.startswith('~'):
        path_to_convert = path_to_convert[1:]
    if path_to_convert.startswith('/'):
        path_to_convert = path_to_convert[1:]
    path_to_convert = '{}{}{}'.format(drive_letter,
                                      ':\\',
                                      path_to_convert).replace('/', '\\')
    return path_to_convert


def column_multiply(arr):
    """
    For any 2d array returns a column vector with the product of the
    columns of each row.

    Parameters
    ----------
    arr : numpy.ndarray

    Returns
    -------
    numpy.ndarray
        A ndarray (column vector) with the products of columns of each
        row of arr as values

    Examples
    --------
    >>> arr = np.arrange(10).reshape(5,2)
    >>> arr
    array([[0, 1],
           [2, 3],
           [4, 5],
           [6, 7],
           [8, 9]])
    >>> column_multiply(arr)
    array([[  0.],
           [  6.],
           [ 20.],
           [ 42.],
           [ 72.]])
    """
    new_arr = ones((arr.shape[0], 1))
    for col in xrange(arr.shape[1]):
        new_arr = new_arr * arr[:, col].reshape(arr.shape[0], 1)
    return new_arr


def stringify(symbol_or_list):
    """
    Returns a list of strings from a list of sympy.Symbol objects or
    a string from a sympy.Symbol.

    Parameters
    ----------
    symbol_or_list : sympy.Symbol or list of sympy.Symbol.

    Returns
    -------
    str or list of str
        A str or list of str representation of the sympy.Symbol or
        list of sympy.Symbol.

    Examples
    --------
    >>> import sympy
    >>> symbol_list = sympy.sympify(['a','c','d'])
    >>> a = stringify(symbol_list[0])
    >>> a
    'a'
    >>> type(a)
    <type 'str>'
    >>> str_list = stringify(symbol_list)
    >>> str_list
    ['a', 'b', 'c']
    >>> type(str_list[0])
    <type 'str>'

    """
    if is_iterable(symbol_or_list):
        return [str(symbol) for symbol in symbol_or_list]
    else:
        return str(symbol_or_list)


def is_iterable(obj):
    """
    Returns True if an object is iterable and False if it is not.

    This function makes the assumtion that any iterable object can be
    cast as an iterator using the build-in function `iter`. This might
    not be the case, but works within the context of PySCeSToolbox.

    Parameters
    ----------
    obj : object
        Any object that might or might not be iterable.

    Returns
    -------
    bool
        A boolean indicating if `ob` is iterable.
    """
    try:
        iter(obj)
        return True
    except TypeError:
        return False


def is_linear(scan_range):
    """
    For any 1-demensional data structure containing numbers return True
    if the numbers follows a linear sequence.

    Within the context of PySCeSToolbox this function will be called on
    either a linear range or a log range. Thus, while not indicative
    of log ranges, this is what a False return value indicates in this
    software.

    Parameters
    ----------
    scan_range : iterable
        Any iterable object containing a range of numbers.

    Returns
    -------
    bool
        A boolean indicating if `scan_range` is a linear sequence of
        numbers.
    """

    if scan_range[1] - scan_range[0] == scan_range[-1] - scan_range[-2]:
        return True
    else:
        return False


def scanner_range_setup(scan_range):
    """
    From a range of numbers, returns its start point, end point, number
    of points and if it is a log range.

    The assumption is made that only log or linear ranges are valid
    inputs, thus lists of random numbers would likely be classified as
    logarithmic.

    Parameters
    ----------
    scan_range : iterable
        Any iterable object containing a range of numbers. Most probably
        numpy.ndarray.

    Returns
    -------
    start : number
        A number indicating the start point of the scan range.
    end: number
        A number indicating the end point of the scan range.
    scan_points: number
        A number indicating the number of scan point in the scan range
    is_log_range: bool
        A boolean indicating whether the scan range has a logarithmic
        scale or not.
    """
    start = scan_range[0]
    end = scan_range[-1]
    scan_points = len(scan_range)
    # based on input not linear == log
    is_log_range = not is_linear(scan_range)
    return start, end, scan_points, is_log_range


def extract_model(obj):
    mod = obj
    try:
        mod = obj.mod
    except:
        pass

    assert isinstance(mod, PysMod), "The object provided does not contain a " \
        "valid Pysces model. Reinstantiate with a" \
        " valid model."

    the_type = str(type(obj))[:-2]
    the_type = the_type[the_type.rindex('.') + 1:]

    return mod, the_type


def find_min(array_like):
    no_zeros = array_like[nonzero(array_like)]
    with errstate(all='ignore'):
        min = nanmin(10 ** log10(no_zeros))
    return min


def find_max(array_like):
    no_zeros = array_like[nonzero(array_like)]
    with errstate(all='ignore'):
        max = nanmax(10 ** log10(no_zeros))
    return max


def do_safe_state(mod, parameter, value, type='ss'):
    setattr(mod, parameter, value)
    mod.SetQuiet()
    mod.doState()
    if mod.__StateOK__:
        if type is 'mca':
            mod.EvalEvar()
            mod.EvalEpar()
            mod.EvalCC()
        ret = True
    else:
        ret = False
    mod.SetLoud()
    return ret


def get_value_eval(expression, subs_dict):
    for k, v in subs_dict.iteritems():
        subs_dict[k] = float64(v)
    ans = eval(expression, {}, subs_dict)
    # sometimes subs_dict is empty
    if not len(subs_dict) == 0:
        # if not, check that if the values of subs_dict are arrays,
        # the answer and the subs_dict arrays are of equal length
        # there is probably a faster solution, but this works fine by
        # falling back to get_value_sympy
        if is_iterable(subs_dict.values()[0]):
            if len(ans) != len(subs_dict.values()[0]):
                raise Exception
    # print 'gve'
    return ans


def get_value_sympy(expression, subs_dict):
    # this is quite inefficient - but probably worth it

    expression = sympify(expression)
    # dont bother with anything if the subs_dict is empty. Just do the
    # "substitution"
    if len(subs_dict) == 0:
        ans = expression.subs(subs_dict)
    else:
        first_value = subs_dict.values()[0]
        if is_iterable(first_value):
            ans = []
            if not isinstance(first_value[0], float64):
                for k, v in subs_dict.iteritems():
                    subs_dict[k] = float64(v)

            number_of_array_elements = len(first_value)
            for i in range(number_of_array_elements):
                temp_subsdict = {k: v[i] for k, v in subs_dict.iteritems()}
                ans_val = expression.subs(temp_subsdict)
                ans.append(ans_val)
            ans = float64(array(ans))
        else:
            if not isinstance(first_value, float64):
                for k, v in subs_dict.iteritems():
                    subs_dict[k] = float64(v)
            ans = expression.subs(subs_dict)
    # print 'gvs'
    return ans


def get_value(expression, subs_dict):
    try:
        # reasons for excepting:
        #    expression is just a single number
        #    expression contains mathematical functions e.g. log, sin, cos
        return get_value_eval(expression, subs_dict)
    except:
        return get_value_sympy(expression, subs_dict)


def split_coefficient(coefficient_name, mod):
    coefficient_2_types = ['cc', 'ec', 'rc']
    coefficient_3_types = ['prc']
    if coefficient_name[:2] in coefficient_2_types:
        coefficient_type = coefficient_name[:2]
    elif coefficient_name[:3] in coefficient_3_types:
        coefficient_type = coefficient_name[:3]
    else:
        return None
    try:
        members = eval(coefficient_type + '_dict(mod)["' + coefficient_name + '"]')
        return_val = tuple([coefficient_type] + list(members))
    except KeyError:
        return_val = None
    return return_val


def is_number(suspected_number):
    """
    Test if an object is a number

    Parameters
    ----------
    suspected_number: object
        This can be any object which might be a number.

    Returns
    -------
    boolean
        True if object is a number, else false

    """

    # We make the assumption that most numbers can be converted to ints.
    # - 21/10/2015 False assumption - strings representing that look like ints
    # can be converted to ints
    # new assumption - only numbers can add another number
    # previous assumption is false - booleans are treated as numbers
    # false many other objects implement addition of numbers
    # original way was fine, we just need to exclude bools and strings.
    # this will not work for invalid types that implement an `__int__` method.
    # but for my cases it should not be a problem.
    number = False
    the_type = type(suspected_number)
    if the_type not in (bool, str, bool_, string_):
        try:
            int(suspected_number)
            # suspected_number + 1
            number = True
        except Exception:
            pass
    return number


def formatter_factory(min_val=None,
                      max_val=None,
                      default_fmt=None,
                      outlier_fmt=None):
    """Returns a custom `html_table` object cell content formatter function.

    Parameters
    ----------
    min_val : int or float, optional (Default : 0.001)
        The minimum value for float display cutoff.
    max_val : int of float, optional (Default : 10000)
        The maximum value for float display cutoff.
    default_fmt : str, options (Default : '%.3f')
        The default format for any number within the range of min_val to
        max_val.
    outlier_fmt : str, optional (Default : '%.3e')
        The format for any number not in the range of min_val to max_val

    Returns
    -------
    formatter : function
        A function which formats input for `html_table` using the values set
        up by this function.

    Examples
    --------
    >>> f = formatter_factory(min_val=1,
                              max_val=10,
                              default_fmt='%.2f',
                              outlier_fmt='%.2e')
    >>> f(1)
    '1.00'
    >>> f(5.235)
    '5.24'
    >>> f(10)
    '10.00'
    >>> f(0.99842)
    '9.98e-01'
    >>> f('abc')
    'abc'

    """
    if not max_val:
        max_val = 10000
    if not min_val:
        min_val = 0.001
    if not default_fmt:
        default_fmt = '%.3f'
    if not outlier_fmt:
        outlier_fmt = '%.3e'

    def formatter(to_format):
        fmt = '%s'
        if is_number(to_format):
            if abs(to_format) == 0:
                fmt = default_fmt
            elif abs(to_format) >= max_val or abs(to_format) < min_val:
                fmt = outlier_fmt
            else:
                fmt = default_fmt
        return fmt % to_format

    return formatter


def html_table(matrix_or_array_like,
               float_fmt=None,
               raw=False,
               first_row_headers=False,
               caption=None,
               style=None,
               formatter=None):
    """Constructs an html compatible table from 2D list, numpy array or sympy
    matrix.

    Parameters
    ----------
    matrix_or_array_like : list of lists or array or matrix
        A compatible object to be converted to an html table
    float_fmt : str, optional (Default : '%.2f')
        The formatter string for numbers. This formatter will be applied to all
        numbers. This optional argument is only used when the argument
        `formatter` is None. Useful for simple tables where different types
        of formatting is not needed.
    raw : boolean, optional (Default : False)
        If True a raw html string will be returned, otherwise an IPython `HTML`
        object will be returned.
    first_row_headers : boolean, optional (Default : False)
        If True elements in the fist row in `matrix_or_array_like` will be
        considered as part of a header and will get the <th></th> tag,
        otherwise there will be no header.
    caption : str, optional (Default : None)
        An optional caption for the table.
    style : str, optional (Default : None)
        An optional html table style
    formatter: function, optional (Default : None)
        An optional `formatter` function. If none float_fmt will be used to
        format numbers.

    Returns
    -------
    str
        A string containing an html table.
    OR
    HTML
        An IPython notebook `HTML` object.


    See Also
    --------
    formatter_factory

    """

    raw_table = matrix_or_array_like
    if not float_fmt:
        float_fmt = '%.2f'

    if not formatter:
        formatter = formatter_factory(default_fmt=float_fmt,
                                      outlier_fmt=float_fmt)

    if 'sympy.matrices' in str(type(matrix_or_array_like)):
        raw_table = array(raw_table)
    if style:
        html_table = ['<table  style="%s">' % style]
    else:
        html_table = ['<table>']
    if caption:
        html_table.append('<caption>%s</caption>' % caption)
    row_count = 0
    for row in raw_table:
        html_table.append('<tr>')
        for col in row:
            to_append = formatter(col)

            if first_row_headers and row_count == 0:
                html_table.append('<th>{0}</th>'.format(to_append))
            else:
                html_table.append('<td>{0}</td>'.format(to_append))

        html_table.append('</tr>')
        row_count += 1
    html_table.append('</table>')
    if raw:
        return ''.join(html_table)
    else:
        return HTML(''.join(html_table))


def silence_print(func):
    """
    A function wrapper that silences the stdout output of a function.

    This function is *very* useful for silencing pysces functions that
    print a lot of unneeded output.

    Parameters
    ----------
    func : function
        A function that talks too much.

    Returns
    -------
    function
        A very quiet function
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        # stdout = sys.stdout
        try:
            sys.stdout = open(devnull, 'w')
            returns = func(*args, **kwargs)
            fix_printing()
            return returns
        except KeyboardInterrupt, e:
            fix_printing()
            raise e
    return wrapper

def fix_printing():
    sys.stdout = ConfigReader.get_config()['stdout']


@memoize
def cc_list(mod):
    """
    Retuns a list of control coefficients of a model.

    The list contains both flux and species control coefficients and control
    coefficients follow the syntax of 'cc_controlled_controller'.

    Parameters
    ----------
    mod : PysMod
        The Pysces model contains the reactions and species which is used to
        construct the control coefficient list.

    Returns
    -------
    list of str
        The cc_list is sorted alphabetically.

    See Also
    --------
    ec_list, rc_list, prc_list

    """
    ccs = []
    for base_reaction in mod.reactions:
        for top_species in mod.species:
            cc = 'cc%s_%s' % (top_species, base_reaction)
            ccs.append(cc)
        for top_reaction in mod.reactions:
            cc = 'ccJ%s_%s' % (top_reaction, base_reaction)
            ccs.append(cc)
    ccs.sort()
    return ccs

@memoize
def cc_dict(mod):
    ccs = {}
    for base_reaction in mod.reactions:
        for top_species in mod.species:
            cc = 'cc%s_%s' % (top_species, base_reaction)
            ccs[cc] = (top_species, base_reaction)
        for top_reaction in mod.reactions:
            cc = 'ccJ%s_%s' % (top_reaction, base_reaction)
            ccs[cc] = (top_reaction, base_reaction)
    return ccs

@memoize
def ec_list(mod):
    """
    Retuns a list of elasticity coefficients of a model.

    The list contains both species and parameter elasticity coefficients and
    elasticity coefficients follow the syntax of 'ec_reaction_sp-or-param'.

    Parameters
    ----------
    mod : PysMod
        The Pysces model contains the reactions, species and parameters
        which is used to construct the elasticity coefficient list.

    Returns
    -------
    list of str
        The ec_list is sorted alphabetically.

    See Also
    --------
    cc_list, rc_list, prc_list

    """
    ecs = []
    for top_reaction in mod.reactions:
        for base_species in mod.species:
            ec = 'ec%s_%s' % (top_reaction, base_species)
            ecs.append(ec)
        for base_param in mod.parameters:
            ec = 'ec%s_%s' % (top_reaction, base_param)
            ecs.append(ec)
    ecs.sort()
    return ecs

@memoize
def ec_dict(mod):
    ecs = {}
    for top_reaction in mod.reactions:
        for base_species in mod.species:
            ec = 'ec%s_%s' % (top_reaction, base_species)
            ecs[ec] = (top_reaction, base_species)
        for base_param in mod.parameters:
            ec = 'ec%s_%s' % (top_reaction, base_param)
            ecs[ec] = (top_reaction, base_param)
    return ecs

@memoize
def rc_list(mod):
    """
    Retuns a list of response coefficients of a model.

    The list contains both species and flux response coefficients and
    response coefficients follow the syntax of 'rc_responder_parameter'.

    Parameters
    ----------
    mod : PysMod
        The Pysces model contains the reactions, species and parameters
        which is used to construct the response coefficient list.

    Returns
    -------
    list of str
        The rc_list is sorted alphabetically.

    See Also
    --------
    cc_list, ec_list, prc_list

    """
    rcs = []
    for base_param in mod.parameters:
        for top_species in mod.species:
            rc = 'rc%s_%s' % (top_species, base_param)
            rcs.append(rc)
        for top_reaction in mod.reactions:
            rc = 'rcJ%s_%s' % (top_reaction, base_param)
            rcs.append(rc)
    rcs.sort()
    return rcs

@memoize
def rc_dict(mod):
    rcs = {}
    for base_param in mod.parameters:
        for top_species in mod.species:
            rc = 'rc%s_%s' % (top_species, base_param)
            rcs[rc] = (top_species, base_param)
        for top_reaction in mod.reactions:
            rc = 'rcJ%s_%s' % (top_reaction, base_param)
            rcs[rc] = (top_reaction, base_param)
    return rcs

@memoize
def prc_list(mod):
    """
    Retuns a list of partial response coefficients of a model.

    The list contains both species and flux partial response coefficients and
    partial response coefficients follow the syntax of
    'prc_responder_parameter_route'.

    Parameters
    ----------
    mod : PysMod
        The Pysces model contains the reactions, species and parameters
        which is used to construct the partial response coefficient list.

    Returns
    -------
    list of str
        The prc_list is sorted alphabetically.

    See Also
    --------
    cc_list, ec_list, rc_list

    """
    prcs = []
    for base_param in mod.parameters:
        for back_reaction in mod.reactions:
            for top_species in mod.species:
                prc = 'prc%s_%s_%s' % (top_species, base_param, back_reaction)
                prcs.append(prc)
            for top_reaction in mod.reactions:
                prc = 'prcJ%s_%s_%s' % (top_reaction,
                                        base_param,
                                        back_reaction)
                prcs.append(prc)
    prcs.sort()
    return prcs

@memoize
def prc_dict(mod):
    prcs = {}
    for base_param in mod.parameters:
        for back_reaction in mod.reactions:
            for top_species in mod.species:
                prc = 'prc%s_%s_%s' % (top_species, base_param, back_reaction)
                prcs[prc] = (top_species, base_param, back_reaction)
            for top_reaction in mod.reactions:
                prc = 'prcJ%s_%s_%s' % (top_reaction,
                                        base_param,
                                        back_reaction)
                prcs[prc] = (top_reaction,
                             base_param,
                             back_reaction)
    return prcs

@memoize
def flux_list(mod):
    fluxes = []
    for reaction in mod.reactions:
        fluxes.append('J_' + reaction)
    return fluxes

@memoize
def ss_species_list(mod):
    ss_species = []
    for species in mod.species:
        ss_species.append(species + '_ss')
    return ss_species

def group_sort(old_list, num_of_groups):
    # Normally scan columns are sorted in a way that they are sorted together
    # in the order they were scanned. This function sorts the lines so that
    # each item in each group is equally spaced away from each other so that
    # they can be assigned colors (using matplotlib's colormap functionality)
    # that differ from each other. This is better than sorting lines before
    # sending the data tp data2d because otherwise button grouping will also be
    # affected by this sorting scheme.
    group_size = len(old_list) / num_of_groups
    first_group = range(0, len(old_list), num_of_groups)
    groups = first_group
    for i in xrange(1, group_size - 1):
        groups = groups + [j + i for j in first_group]
    new_list = [old_list[pos] for pos in groups]

    return new_list

def print_f(message, status):
    """
    Prints a message if status is `True`
    Parameters
    ----------
    message : object
        Any object with a `__str__` method.
    status : bool
    """
    if status:
        print message


class PseudoDotDict:
    """
    A class that acts like a dictionary with dot accessible elements.

    This class is not subsclassed from ``dict`` like DotDict, but rather wraps
    dictionary functionality.

    This object has trouble being pickled :'(

    See Also
    --------
    dict
    DotDict
    """
    _reserved = ['__class__', '__format__', '__init__', '__reduce_ex__',
                 '__sizeof__', '__delattr__', '__getattribute__', '__new__',
                 '__repr__', '__str__', '__doc__', '__hash__', '__reduce__',
                 '__setattr__', '__subclasshook__', 'keys', 'items', 'values']

    def __init__(self, *args, **kwargs):
        self._dict = dict(*args, **kwargs)
        self._setall_init
        self.keys = self._dict.keys
        self.values = self._dict.values
        self.items = self._dict.items

    def _setall_init(self):
        """
        Internal function that populates the self namespace on
        initialisation. Throws exception if any of the keys are
        reserved.
        """
        for k, v in self._dict.iteritems():
            if k in PseudoDotDict._reserved:
                raise Exception('%s is a reserved key' % k)
            else:
                setattr(self, k, v)

    def __setitem__(self, x, y):
        self._dict[x] = y
        if x in PseudoDotDict._reserved:
            raise Exception('%s is a reserved key' % x)
        else:
            setattr(self, x, y)

    def __getitem__(self, x):
        return self._dict[x]

    def __repr__(self):
        return self._dict.__repr__()

    def update(self, dic):
        for k, v in dic.iteritems():
            self.__setitem__(k, v)


class DotDict(dict):
    """A class that inherits from ``dict``.

    The DotDict class has the same functionality as ``dict`` but with the added
    feature that dictionary elements may be accessed via dot notation.

    See Also
    --------
    dict
    PseudoDotDict
    """

    _reserved1 = ['clear', 'fromkeys', 'has_key', 'iteritems', 'itervalues',
                  'pop', 'setdefault', 'values', 'viewkeys', 'copy', 'get',
                  'items', 'iterkeys', 'keys', 'popitem', 'update',
                  'viewitems', 'viewvalues']

    _reserved2 = ['__class__', '__delitem__', '__ge__', '__hash__', '__len__',
                  '__reduce__', '__setitem__', '__cmp__', '__doc__',
                  '__getattribute__', '__init__', '__lt__', '__reduce_ex__',
                  '__sizeof__', '__contains__', '__eq__', '__getitem__',
                  '__iter__', '__ne__', '__repr__', '__str__', '__delattr__',
                  '__format__', '__gt__', '__le__', '__new__',
                  '__setattr__', '__subclasshook__']

    _reserved = _reserved1 + _reserved2

    def __init__(self, *args, **kwargs):
        super(DotDict, self).__init__(*args, **kwargs)
        self._setall_init()
        # self._reassign_reserved_to_internals()

    def __setitem__(self, x, y):
        dict.__setitem__(self, x, y)
        if x in DotDict._reserved:
            raise Exception
        else:
            setattr(self, x, y)

    def _setall_init(self):
        """
        Internal method that populates the self namespace on
        initialisation. Throws exception if any of the keys are
        reserved.
        """
        for k, v in self.iteritems():
            if k in DotDict._reserved:
                raise Exception('%s is a reserved key' % k)
            else:
                setattr(self, k, v)

    def update(self, dic):
        for k, v in dic.iteritems():
            self.__setitem__(k, v)

    def _make_repr(self, key, value, formatter=None):
        """
        Internal method makes a _repr_html_ method and attaches it to self so
        that values contained in self can be displayed as an html table in the
        IPython notebook.

        Parameters
        ----------
        key : str
            A string that will be evaluated indicating how to represent the
            dictionary keys. If 'k' is passed, the key will be displayed as is.
        value : str
            A string that will be evaluated indicating how to represent
            dictionary values. If 'v' is passed, the value will be displayed as
            is.
        formatter : function, optional (Default : None)
            A formatter function that formats numbers. If none, the default
            function produced by `formatter_factory` will be used.

        See Also
        --------
        formatter_factory


        """
        if not formatter:
            formatter = formatter_factory()

        def representation(the_self=self):
            keys = self.keys()
            keys.sort()
            values = [self[the_key] for the_key in keys]
            items = zip(keys, values)
            lst = []
            for k, v in items:
                col1 = eval(key)
                col2 = eval(value)
                lst.append((col1, col2))

            tables = []
            cur_list = []
            for i, each in enumerate(lst):
                if (i + 1) % 10 == 0:
                    cur_list.append(each)
                    tables.append(
                        html_table(cur_list,
                                   style='display: inline-table',
                                   raw=True,
                                   formatter=formatter))
                    cur_list = []
                elif (i + 1) == len(lst):
                    # print 'Final'
                    cur_list.append(each)
                    tables.append(
                        html_table(cur_list,
                                   style='display: inline-table',
                                   raw=True,
                                   formatter=formatter))
                    cur_list = []
                else:
                    cur_list.append(each)

            div_string = '<div>'
            for each in tables:
                div_string += each
                div_string += '\t\t'
            div_string += '</div>'
            return div_string

        self._repr_html_ = representation


# no use yet for this class but seemed like a nice/quick/simple way to address
# directory trees with no overlapping subdirectory names.
# will work for ['/a/b/c/d',/a/b/c/e'] but not for ['/a/b/c/d',/a/b/d']
class SimpleDirectoryStructure(object):
    def __init__(self, fpaths=None):
        self._path_dict = {}
        if type(fpaths) is list:
            self.add_paths(fpaths)
        if type(fpaths) is str:
            self.add_path(fpaths)

    def add_path(self, fpath):
        top_path = fpath
        temp_split = path.split(top_path)
        self._path_dict[temp_split[1]] = top_path
        while (temp_split[1] != ''):
            top_path = temp_split[0]
            temp_split = path.split(top_path)
            self._path_dict[temp_split[1]] = top_path

    def add_paths(self, list_of_fpaths):
        for fpath in list_of_fpaths:
            self.add_path(fpath)

    def __getitem__(self, key):
        return self._path_dict[key]



from . import modeltools
from . import latextools
from . import utils
from . import analyse
from .analyse import Symca
from .analyse import RateChar
from .analyse import ThermoKin
from .utils.model_graph import ModelGraph
from .utils.plotting import SimpleData2D
from .utils.plotting import Data2D


from .widget import ForceDirectedGraph
from .eventful_graph import EventfulGraph, empty_eventfulgraph_hook

def _jupyter_nbextension_paths():
    return [{
        'section': 'notebook',
        'src': '.',
        'dest': 'd3networkx_psctb',
        'require': 'd3networkx_psctb/widget'
}]


from os import path

from numpy import log10, array, float, NaN, nanmin, nanmax, savetxt, hstack, float
from pysces import ModelMap, ParScanner, Scanner
from sympy import sympify, diff, Symbol

from ._thermokin_file_tools import get_subs_dict, get_reqn_path, \
    get_all_terms, get_term_types_from_raw_data, create_reqn_data, \
    write_reqn_file, create_gamma_keq_reqn_data, term_to_file
from ..latextools import LatexExpr
from ..modeltools import make_path, get_file_path
from ..utils.misc import do_safe_state, get_value, silence_print, print_f, \
    is_number, stringify, scanner_range_setup, DotDict, formatter_factory, \
    find_min, find_max
from ..utils.plotting import Data2D

__author__ = 'carl'
__all__ = ['ThermoKin']


def mult(lst):
    """
    Multiplies values of a list with each other and returns the result.

    Parameters
    ----------
    lst : list of numbers

    Returns
    -------
    number
        Same type as numbers in ``lst``.

    """
    ans = 1
    for each in lst:
        ans *= each
    return ans


def get_repr_latex(obj):
    """
    Creates the string that will be returned by the ``__repr_latex__``
    method of any of objects of ``Thermokin``. The value of the
    ``value`` field is used to dermine the float format.

    Parameters
    ----------
    obj : RateTerm, Term or RateEqn

    Returns
    -------
    str

    """
    if obj.value == 0:
        fmt = '$%s = %s = %.3f$'
    elif abs(obj.value) < 0.001 or abs(obj.value) > 10000:
        fmt = '$%s = %s = %.3e$'
    else:
        fmt = '$%s = %s = %.3f$'
    return fmt % (obj.latex_name,
                  obj.latex_expression,
                  obj.value)


@silence_print
def silent_state(mod):
    mod.doMca()
    mod.doState()


class ThermoKin(object):
    def __init__(self, mod, path_to_reqn_file=None, overwrite=False,
                 warnings=True, ltxe=None):
        super(ThermoKin, self).__init__()
        self.mod = mod

        silent_state(mod)

        self._analysis_method = 'thermokin'
        self._working_dir = make_path(self.mod, self._analysis_method)

        if ltxe:
            self._ltxe = ltxe
        else:
            self._ltxe = LatexExpr(mod)

        if path_to_reqn_file:
            self._path_to = path_to_reqn_file
        else:
            self._path_to = get_reqn_path(self.mod)

        self._do_auto_actions(overwrite, warnings)
        self._raw_data, self._add_raw_data = get_all_terms(self._path_to)
        self._do_gamma_keq(overwrite, warnings)

        term_types = get_term_types_from_raw_data(self._raw_data).union(
            get_term_types_from_raw_data(self._add_raw_data))
        self._ltxe.add_term_types(term_types)

        self._populate_object()
        self._populate_ec_results()

    def _do_gamma_keq(self, overwrite, warnings):
        if overwrite:
            return None
        gamma_keq_todo = []
        add_raw_data = self._add_raw_data
        for reaction in self._raw_data.iterkeys():
            if not add_raw_data.get(reaction) or not add_raw_data.get(
                    reaction).get('gamma_keq'):
                gamma_keq_todo.append(reaction)
        if len(gamma_keq_todo) != 0:
            reaction_printout = ', '.join(gamma_keq_todo[:-1]) + ' or ' + \
                                gamma_keq_todo[-1]
            print_f('%s does not contain Gamma/Keq terms for %s:' % (
                self._path_to, reaction_printout), warnings)
            gamma_keq_data, messages = create_gamma_keq_reqn_data(self.mod)
            for required in gamma_keq_todo:
                print_f('{:10.10}: {}'.format(required, messages[required]),
                        warnings)
                if required not in add_raw_data:
                    add_raw_data[required] = {}
                add_raw_data[required]['gamma_keq'] = gamma_keq_data[required]

    def _do_auto_actions(self, overwrite, warnings):
        condition_1 = path.exists(self._path_to) and overwrite
        condition_2 = not path.exists(self._path_to)
        if condition_1:
            print_f(
                'The file %s will be overwritten with automatically generated file.' % self._path_to,
                warnings)
        elif condition_2:
            print_f('A new file will be created at "%s".' % self._path_to,
                    warnings)
        if condition_1 or condition_2:
            ma_terms, vc_binding_terms, gamma_keq_terms, messages = create_reqn_data(
                self.mod)
            for k, v in messages.iteritems():
                print_f('{:10.10}: {}'.format(k, v), warnings)
            write_reqn_file(self._path_to, self.mod.ModelFile, ma_terms,
                            vc_binding_terms, gamma_keq_terms, messages)

    def _populate_object(self):
        self.reaction_results = DotDict()
        self.reaction_results._make_repr('"$" + v.latex_name + "$"', 'v.value',
                                         formatter_factory())
        for reaction, terms_dict in self._raw_data.iteritems():
            additional_terms = self._add_raw_data.get(reaction)
            reqn_obj = RateEqn(self.mod,
                               reaction,
                               terms_dict,
                               self._ltxe,
                               additional_terms)
            setattr(self, 'J_' + reaction, reqn_obj)
            self.reaction_results['J_' + reaction] = reqn_obj
            for term in reqn_obj.terms.itervalues():
                self.reaction_results[term.name] = term

    def _populate_ec_results(self):
        self.ec_results = DotDict()
        self.ec_results._make_repr('"$" + v.latex_name + "$"', 'v.value',
                                   formatter_factory())

        for rate_eqn in self.reaction_results.itervalues():
            self.ec_results.update(rate_eqn.ec_results)

    def save_results(self, file_name=None, separator=',',fmt='%.9f'):
        file_name = get_file_path(working_dir=self._working_dir,
                                  internal_filename='tk_summary',
                                  fmt='csv',
                                  file_name=file_name, )

        values = []
        max_len = 0
        for reaction_name in sorted(self.reaction_results.keys()):
            cols = (reaction_name,
                    self.reaction_results[reaction_name].value,
                    self.reaction_results[reaction_name].latex_name,
                    self.reaction_results[reaction_name].latex_expression)
            values.append(cols)
            if len(cols[3]) > max_len:
                max_len = len(cols[3])

        for elasticity_name in sorted(
                [ec for ec in self.ec_results.keys() if ec.startswith('ec')]):
            if self.ec_results[elasticity_name].expression != 0:
                related_ecs = sorted([ec for ec in self.ec_results.keys() if
                                      elasticity_name in ec])
                for related_ec_name in related_ecs:
                    cols = (related_ec_name,
                            self.ec_results[related_ec_name].value,
                            self.ec_results[related_ec_name].latex_name,
                            self.ec_results[related_ec_name].latex_expression)
                    values.append(cols)
                    if len(cols[3]) > max_len:
                        max_len = len(cols[3])

        str_fmt = 'S%s' % max_len
        head = ['name', 'value', 'latex_name', 'latex_expression']
        X = array(values,
                  dtype=[(head[0], str_fmt),
                         (head[1], 'float'),
                         (head[2], str_fmt),
                         (head[3], str_fmt)])

        try:
            savetxt(fname=file_name,
                    X=X,
                    header=separator.join(head),
                    delimiter=separator,
                    fmt=['%s', fmt, '%s', '%s'], )

        except IOError as e:
            print e.strerror


class RateEqn(object):
    def __init__(self, mod, name, term_dict, ltxe, additional_terms=None):
        super(RateEqn, self).__init__()
        self.mod = mod
        self.terms = DotDict()
        self.terms._make_repr('"$" + v.latex_name + "$"', 'v.value',
                              formatter_factory())
        self._unfac_expression = 1
        self.name = 'J_' + name
        self._rname = name
        self._ltxe = ltxe

        for val in term_dict.itervalues():
            self._unfac_expression = self._unfac_expression * (sympify(val))
        for term_name, expression in term_dict.iteritems():
            term = RateTerm(parent=self,
                            mod=self.mod,
                            name='J_%s_%s' % (self._rname, term_name),
                            rname=term_name,
                            expression=expression,
                            ltxe=self._ltxe)
            setattr(self, term_name, term)
            self.terms[term_name] = term

        if additional_terms:
            for term_name, expression in additional_terms.iteritems():
                term = AdditionalRateTerm(parent=self,
                                          mod=self.mod,
                                          name='J_%s_%s' % (
                                              self._rname, term_name),
                                          rname=term_name,
                                          expression=expression,
                                          ltxe=self._ltxe)
                setattr(self, term_name, term)
                self.terms[term_name] = term

        self._value = None
        self._str_expression_ = None
        self._expression = None
        self._latex_expression = None
        self._latex_name = None

        self.ec_results = DotDict()
        self.ec_results._make_repr('"$" + v.latex_name + "$"', 'v.value',
                                   formatter_factory())

        self._populate_ec_results()

    def _populate_ec_results(self):
        expression_symbols = self._unfac_expression.atoms(Symbol)
        for each in expression_symbols:
            each = sympify(each)
            ec = diff(self._unfac_expression, each) * \
                (each / self._unfac_expression)
            ec_name = 'ec%s_%s' % (self._rname, each)
            self.ec_results[ec_name] = Term(self, self.mod, ec_name,
                                            self._rname, ec,
                                            self._ltxe)
        for each in self.terms.itervalues():
            self.ec_results.update(each.ec_results)

    def _repr_latex_(self):
        return get_repr_latex(self)

    @property
    def _str_expression(self):
        if not self._str_expression_:
            self._str_expression_ = str(self._unfac_expression)
        return self._str_expression_

    @property
    def expression(self):
        if not self._expression:
            self._expression = self._unfac_expression.factor()
        return self._expression

    @property
    def value(self):
        self._calc_value()
        return self._value

    @property
    def latex_name(self):
        if not self._latex_name:
            self._latex_name = self._ltxe.expression_to_latex(self.name)
        return self._latex_name

    @property
    def latex_expression(self):
        if not self._latex_expression:
            self._latex_expression = self._ltxe.expression_to_latex(
                self.expression,
                mul_symbol='dot')
        return self._latex_expression

    def _calc_value(self):
        subs_dict = get_subs_dict(self._unfac_expression, self.mod)
        for each in self.terms.itervalues():
            if type(each) is not AdditionalRateTerm:
                each._calc_value(subs_dict)
        self._value = mult([each._value for each in self.terms.itervalues() if
                            type(each) is not AdditionalRateTerm])

    def _valscan_x(self, parameter, scan_range):
        scan_res = [list() for _ in range(len(self.terms.values()) + 2)]
        scan_res[0] = scan_range

        for parvalue in scan_range:
            state_valid = do_safe_state(self.mod, parameter, parvalue)
            for i, term in enumerate(self.terms.values()):
                if state_valid:
                    scan_res[i + 1].append(term.value)
                else:
                    scan_res[i + 1].append(NaN)
            if state_valid:
                scan_res[i + 2].append(self.value)
            else:
                scan_res[i + 2].append(NaN)
        return scan_res

    def _valscan(self,
                 parameter,
                 scan_range,
                 par_scan=False,
                 par_engine='multiproc'):

        # choose between parscanner or scanner
        if par_scan:
            # This is experimental
            scanner = ParScanner(self.mod, par_engine)
        else:
            scanner = Scanner(self.mod)
            scanner.quietRun = True

        # parameter scan setup and execution
        start, end, points, log = scanner_range_setup(scan_range)
        scanner.addScanParameter(parameter,
                                 start=start,
                                 end=end,
                                 points=points,
                                 log=log)

        needed_symbols = [parameter] + \
            stringify(list(self.expression.atoms(Symbol)))

        scanner.addUserOutput(*needed_symbols)
        scanner.Run()

        # getting term/reaction values via substitution
        subs_dict = {}
        for i, symbol in enumerate(scanner.UserOutputList):
            subs_dict[symbol] = scanner.UserOutputResults[:, i]

        term_expressions = [term.expression for term in self.terms.values()]\
            + [self.expression]
        term_str_expressions = stringify(term_expressions)
        parameter_values = subs_dict[parameter].reshape(points, 1)

        scan_res = []

        # collecting results in an array
        for expr in term_str_expressions:
            scan_res.append(get_value(expr, subs_dict))
        scan_res = array(scan_res).transpose()
        scan_res = hstack([parameter_values, scan_res])
        return scan_res

    def _ecscan(self,
                parameter,
                scan_range,
                par_scan=False,
                par_engine='multiproc'):

        # choose between parscanner or scanner
        if par_scan:
            # This is experimental
            scanner = ParScanner(self.mod, par_engine)
        else:
            scanner = Scanner(self.mod)
            scanner.quietRun = True

        # parameter scan setup and execution
        start, end, points, log = scanner_range_setup(scan_range)
        scanner.addScanParameter(parameter,
                                 start=start,
                                 end=end,
                                 points=points,
                                 log=log)

        needed_symbols = [parameter] + \
            stringify(list(self.expression.atoms(Symbol)))

        scanner.addUserOutput(*needed_symbols)
        scanner.Run()

        # getting term/reaction values via substitution
        subs_dict = {}
        for i, symbol in enumerate(scanner.UserOutputList):
            subs_dict[symbol] = scanner.UserOutputResults[:, i]

        # we include all ec_terms that are not zero (even though they are
        # included in the main dict)
        ec_term_expressions = [ec_term.expression for ec_term in
                               self.ec_results.values() if
                               ec_term.expression != 0 and
                               not ec_term.name.endswith('gamma_keq')]
        ec_term_str_expressions = stringify(ec_term_expressions)
        parameter_values = subs_dict[parameter].reshape(points, 1)

        scan_res = []

        # collecting results in an array
        for expr in ec_term_str_expressions:
            val = get_value(expr, subs_dict)
            scan_res.append(val)
        scan_res = array(scan_res).transpose()
        scan_res = hstack([parameter_values, scan_res])
        return scan_res

    def _ecscan_x(self, parameter, scan_range):
        mca_objects = [ec_term for ec_term in self.ec_results.values() if
                       ec_term.expression != 0 and not ec_term.name.endswith(
                           'gamma_keq')]

        scan_res = [list() for _ in range(len(mca_objects) + 1)]
        scan_res[0] = scan_range

        for parvalue in scan_range:
            state_valid = do_safe_state(self.mod, parameter, parvalue,
                                        type='mca')
            for i, term in enumerate(mca_objects):
                if state_valid:
                    scan_res[i + 1].append(term.value)
                else:
                    scan_res[i + 1].append(NaN)
        return scan_res

    def do_par_scan(self,
                    parameter,
                    scan_range,
                    scan_type='value',
                    init_return=True,
                    par_scan=False,
                    par_engine='multiproc'):

        try:
            assert scan_type in ['elasticity', 'value'], 'scan_type must be one\
                of "value" or "elasticity".'
        except AssertionError as ae:
            print ae

        init = getattr(self.mod, parameter)

        if scan_type == 'elasticity':
            mca_objects = [ec_term for ec_term in self.ec_results.values() if
                           ec_term.expression != 0 and
                           not ec_term.name.endswith('gamma_keq')]

            additional_cat_classes = {
                'All Coefficients': ['Term Elasticities']}
            additional_cats = {
                'Term Elasticities': [ec_term.name for ec_term in mca_objects
                                      if
                                      ec_term.name.startswith('p')]}
            column_names = [parameter] + \
                [ec_term.name for ec_term in mca_objects]
            y_label = 'Elasticity Coefficient'
            scan_res = self._ecscan(parameter,
                                    scan_range,
                                    par_scan,
                                    par_engine)
            data_array = scan_res
            # ylim = [nanmin(data_array[:, 1:]),
            #         nanmax(data_array[:, 1:]) * 1.1]
            yscale = 'linear'
            category_manifest = {pec: True for pec in
                                 additional_cats['Term Elasticities']}
            category_manifest['Elasticity Coefficients'] = True
            category_manifest['Term Elasticities'] = True

        elif scan_type == 'value':
            additional_cat_classes = {'All Fluxes/Reactions/Species':
                                      ['Term Rates']}
            term_names = [term.name for term in self.terms.values()]
            additional_cats = {'Term Rates': term_names}
            column_names = [parameter] + term_names + [self.name]
            y_label = 'Reaction/Term rate'
            scan_res = self._valscan(parameter,
                                     scan_range,
                                     par_scan,
                                     par_engine)
            data_array = scan_res
            # ylim = [nanmin(data_array[:, 1:]),
            #         nanmax(data_array[:, 1:]) * 1.1]
            yscale = 'log'
            category_manifest = {'Flux Rates': True, 'Term Rates': True}

        if init_return:
            self.mod.SetQuiet()
            setattr(self.mod, parameter, init)
            self.mod.doMca()
            self.mod.SetLoud()

        mm = ModelMap(self.mod)
        species = mm.hasSpecies()
        if parameter in species:
            x_label = '[%s]' % parameter.replace('_', ' ')
        else:
            x_label = parameter

        xscale = 'log' if scanner_range_setup(scan_range)[3] else 'linear'
        ax_properties = {'ylabel': y_label,
                         'xlabel': x_label,
                         'xscale': xscale,
                         'yscale': yscale, }

        data = Data2D(mod=self.mod,
                      column_names=column_names,
                      data_array=data_array,
                      ltxe=self._ltxe,
                      analysis_method='thermokin',
                      ax_properties=ax_properties,
                      additional_cat_classes=additional_cat_classes,
                      additional_cats=additional_cats,
                      category_manifest=category_manifest,)

        if scan_type == 'elasticity':
            ec_names = [ec_term.name for ec_term in mca_objects if
                        ec_term.name.startswith('ec')]
            for line in data._lines:
                for ec_name in ec_names:
                    condition1 = line.name != ec_name
                    condition2 = self.ec_results[line.name]._rname == ec_name
                    if condition1 and condition2:
                        line.categories.append(ec_name)

        return data

    def __add__(self, other):
        return generic_term_operation(self, other, '+')

    def __mul__(self, other):
        return generic_term_operation(self, other, '*')

    def __sub__(self, other):
        return generic_term_operation(self, other, '-')

    def __div__(self, other):
        return generic_term_operation(self, other, '/')

    def __radd__(self, other):
        return generic_term_operation(self, other, '+')

    def __rmul__(self, other):
        return generic_term_operation(self, other, '*')

    def __rsub__(self, other):
        return generic_term_operation(self, other, 'rsub')

    def __rdiv__(self, other):
        return generic_term_operation(self, other, 'rdiv')

    def __neg__(self):
        return AdditionalTerm(self,
                              self.mod,
                              '-' + self.name,
                              '-' + self._rname,
                              (-self._unfac_expression),
                              self._ltxe,
                              '-' + self.name)

    def __pow__(self, power, modulo=None):
        return AdditionalTerm(self,
                              self.mod,
                              self.name + '**' + str(power),
                              self._rname + '**' + str(power),
                              self._unfac_expression ** power,
                              self._ltxe,
                              self.name + '**' + str(power))


class Term(object):
    def __init__(self, parent, mod, name, rname, expression, ltxe):
        super(Term, self).__init__()
        self.name = name
        self._rname = rname
        self._unfac_expression = sympify(expression)
        self._parent = parent
        self.mod = mod
        self._ltxe = ltxe
        # properties
        self._expression = None
        self._str_expression_ = None
        self._value = None
        self._latex_name = None
        self._latex_expression = None

    def _repr_latex_(self):
        return get_repr_latex(self)

    @property
    def _str_expression(self):
        if not self._str_expression_:
            self._str_expression_ = str(self._unfac_expression)
        return self._str_expression_

    @property
    def expression(self):
        if not self._expression:
            self._expression = self._unfac_expression.factor()
        return self._expression

    @property
    def value(self):
        self._calc_value()
        return self._value

    @property
    def latex_name(self):
        if not self._latex_name:
            self._latex_name = self._ltxe.expression_to_latex(self.name)
        return self._latex_name

    @property
    def latex_expression(self):
        if not self._latex_expression:
            self._latex_expression = self._ltxe.expression_to_latex(
                self.expression,
                mul_symbol='dot')
        return self._latex_expression

    def _calc_value(self, subs_dict=None):
        if not subs_dict:
            subs_dict = get_subs_dict(self._unfac_expression, self.mod)
        self._value = get_value(self._str_expression, subs_dict)

    def __add__(self, other):
        return generic_term_operation(self, other, '+')

    def __mul__(self, other):
        return generic_term_operation(self, other, '*')

    def __sub__(self, other):
        return generic_term_operation(self, other, '-')

    def __div__(self, other):
        return generic_term_operation(self, other, '/')

    def __radd__(self, other):
        return generic_term_operation(self, other, '+')

    def __rmul__(self, other):
        return generic_term_operation(self, other, '*')

    def __rsub__(self, other):
        return generic_term_operation(self, other, 'rsub')

    def __rdiv__(self, other):
        return generic_term_operation(self, other, 'rdiv')

    def __neg__(self):
        return AdditionalTerm(self._parent,
                              self.mod,
                              '-' + self.name,
                              '-' + self._rname,
                              (-self._unfac_expression),
                              self._ltxe,
                              '-' + self.name)

    def __pow__(self, power, modulo=None):
        return AdditionalTerm(self._parent,
                              self.mod,
                              self.name + '**' + str(power),
                              self._rname + '**' + str(power),
                              self._unfac_expression ** power,
                              self._ltxe,
                              self.name + '**' + str(power))


class RateTerm(Term):
    def __init__(self, parent, mod, name, rname, expression, ltxe):
        super(RateTerm, self).__init__(parent, mod, name, rname, expression,
                                       ltxe)
        self.ec_results = DotDict()
        self.ec_results._make_repr('"$" + v.latex_name + "$"', 'v.value',
                                   formatter_factory())
        self._populate_ec_results()
        self._percentage = None

    @property
    def percentage(self):
        per = (log10(self.value) / log10(self._parent.value)) * 100
        return per

    def _populate_ec_results(self):
        expression_symbols = self._parent._unfac_expression.atoms(Symbol)
        expression_symbols.update(self._unfac_expression.atoms(Symbol))
        for each in expression_symbols:
            each = sympify(each)
            ec_name = 'ec%s_%s' % (self._parent._rname, each)
            pec_name = 'p%s_%s' % (ec_name, self._rname)
            ec = diff(self._unfac_expression, each) * \
                (each / self._unfac_expression)
            self.ec_results[pec_name] = Term(self._parent,
                                             self.mod,
                                             pec_name,
                                             ec_name,
                                             ec,
                                             self._ltxe)


class AdditionalRateTerm(RateTerm):
    @property
    def percentage(self):
        return 0.0

    def append_to_file(self, file_name, term_name=None, parent=None):
        if not parent:
            parent = self._parent._rname
        if not term_name:
            term_name = self._rname
        term_to_file(file_name, self._unfac_expression, parent, term_name)


class AdditionalTerm(Term):
    def __init__(self, parent, mod, name, rname, expression, ltxe,
                 creation_operation):
        super(AdditionalTerm, self).__init__(parent, mod, name, rname,
                                             expression,
                                             ltxe)

        self.creation_operation = creation_operation

    def simplify_expression(self):
        self._expression = self._unfac_expression.factor()
        self._latex_expression = None

    def get_elasticity(self, var_par, term_name=None):
        if not term_name:
            term_name = self.name

        var_par = sympify(var_par)
        ec = diff(self._unfac_expression, var_par) * \
            (var_par / self._unfac_expression)
        ec_name = 'ec_%s_%s' % (term_name, var_par)
        ec_term = Term(self, self.mod, ec_name, ec_name, ec, self._ltxe)
        ec_term._latex_name = '\\varepsilon^{%s}_{%s}' % (
            term_name.replace('_', ''),
            str(var_par).replace('_', ''))
        return ec_term

    def append_to_file(self, file_name, term_name=None, parent=None):
        if not parent and self._parent:
            parent = self._parent._rname
        if not term_name and self.name != 'new_term':
            term_name = self.name
        term_to_file(file_name, self._unfac_expression, parent, term_name)

    @property
    def expression(self):
        if not self._expression:
            self._expression = self._unfac_expression
        return self._expression


def generic_term_operation(self, other, operator, parent=None, name=None,
                           rname=None):
    def get_parent(self):
        if type(self) is RateEqn:
            parent = self
        else:
            parent = self._parent
        return parent

    if operator == 'rsub':
        self = -self
        operator = '+'

    if operator == 'rdiv':
        self = AdditionalTerm(get_parent(self),
                              self.mod,
                              self.name,
                              self._rname,
                              (1 / self._unfac_expression),
                              self._ltxe,
                              '1/' + self.name)
        operator = '*'

    if is_number(other):
        other = AdditionalTerm(get_parent(self),
                               self.mod,
                               str(other),
                               str(other),
                               sympify(other),
                               self._ltxe,
                               str(other))

    # TODO this type check is a hack - no idea how to check specifically for
    # sympy expressions
    elif 'sympy' in str(type(other)):
        other = AdditionalTerm(get_parent(self),
                               self.mod,
                               str(other),
                               str(other),
                               other,
                               self._ltxe,
                               str(other))

    mod = self.mod
    operated_on = []
    for term in (self, other):
        if type(term) is AdditionalTerm:
            operated_on.append(term.creation_operation)
        else:
            operated_on.append(term.name)

    if not name:
        name = 'new_term'
    if not rname:
        rname = 'new_term'

    if not parent:
        if hasattr(self, '_parent') and hasattr(other, '_parent'):
            parent = self._parent
        elif type(self) is RateEqn and type(other) is not RateEqn:
            parent = self
        elif type(other) is RateEqn and type(self) is not RateEqn:
            parent = other

    creation_operation = sympify(
        '%s %s %s' % (operated_on[0], operator, operated_on[1]))
    expression = sympify('(%s) %s (%s)' % (
        self._str_expression, operator, other._str_expression))
    ltxe = self._ltxe
    return AdditionalTerm(parent, mod, name, rname, expression, ltxe,
                          creation_operation)



import ipywidgets as widgets # Widget definitions
from traitlets import Unicode, CInt, CFloat, Bool # Import the base Widget class and the traitlets Unicode class.

# Define our ForceDirectedGraph and its target model and default view.
class ForceDirectedGraph(widgets.DOMWidget):
    _view_module = Unicode('nbextensions/d3networkx_psctb/widget', sync=True)
    _view_name = Unicode('D3ForceDirectedGraphView', sync=True)

    width = CInt(500).tag(sync=True)
    height = CInt(500).tag(sync=True)
    charge = CFloat(270.).tag(sync=True)
    distance = CInt(30.).tag(sync=True)
    strength = CInt(0.3).tag(sync=True)
    gnode_json = Unicode(sync=True)
    straight_links = Bool(False).tag(sync=True)
    show_color_legend = Bool(False).tag(sync=True)

    svg_image = Unicode().tag(sync=True)

    def __init__(self, eventful_graph, *pargs, **kwargs):
        widgets.DOMWidget.__init__(self, *pargs, **kwargs)

        self._eventful_graph = eventful_graph
        self._send_dict_changes(eventful_graph.graph, 'graph')
        self._send_dict_changes(eventful_graph.node, 'node')
        self._send_dict_changes(eventful_graph.adj, 'adj')

    # def _dump_svg(self):
    #     self.send({'dict':'save_svg', 'action':None, 'key':None})

    def _ipython_display_(self, *pargs, **kwargs):

        # Show the widget, then send the current state
        widgets.DOMWidget._ipython_display_(self, *pargs, **kwargs)
        for (key, value) in self._eventful_graph.graph.items():
            self.send({'dict': 'graph', 'action': 'add', 'key': key, 'value': value})
        for (key, value) in self._eventful_graph.node.items():
            self.send({'dict': 'node', 'action': 'add', 'key': key, 'value': value})
        for (key, value) in self._eventful_graph.adj.items():
            self.send({'dict': 'adj', 'action': 'add', 'key': key, 'value': value})

    def _send_dict_changes(self, eventful_dict, dict_name):
        def key_add(key, value):
            self.send({'dict': dict_name, 'action': 'add', 'key': key, 'value': value})
        def key_set(key, value):
            self.send({'dict': dict_name, 'action': 'set', 'key': key, 'value': value})
        def key_del(key):
            self.send({'dict': dict_name, 'action': 'del', 'key': key})
        eventful_dict.on_add(key_add)
        eventful_dict.on_set(key_set)
        eventful_dict.on_del(key_del)
